from typing import Any, Dict, Generic, List, Optional, Set, Type, Union

from django.db.models import Model
from ninja.pagination import PaginationBase
from pydantic import BaseModel as PydanticModel
from pydantic import Field, field_validator

try:
    from ninja_schema.errors import ConfigError
    from ninja_schema.orm.factory import SchemaFactory
    from ninja_schema.orm.model_schema import (
        ModelSchemaConfig as NinjaSchemaModelSchemaConfig,
    )
    from ninja_schema.orm.model_schema import (
        ModelSchemaConfigAdapter,
    )
except Exception:  # pragma: no cover
    ConfigError = NinjaSchemaModelSchemaConfig = ModelSchemaConfigAdapter = (
        SchemaFactory
    ) = None


from ...pagination import PageNumberPaginationExtra, PaginatedResponseSchema


class ModelPagination(PydanticModel):
    """
    Model Controller Pagination Configuration
    """

    klass: Type[PaginationBase] = PageNumberPaginationExtra
    paginator_kwargs: Optional[dict] = None
    pagination_schema: Type[PydanticModel] = PaginatedResponseSchema

    @field_validator("pagination_schema", mode="before")
    def validate_schema(cls, value: Any) -> Any:
        if (
            isinstance(value, type)
            and issubclass(value, PydanticModel)
            and issubclass(value, Generic)  # type:ignore[arg-type]
        ):
            return value
        raise ValueError(
            f"{value} is not a valid type. Please use a generic pydantic model."
        )


class ModelSchemaConfig(PydanticModel):
    """
    Model Controller Auto Schema Generation Configuration
    """

    include: Union[str, List[str]] = Field(default="__all__")
    exclude: Set[str] = Field(set())
    optional: Optional[Union[str, Set[str]]] = Field(default=None)
    depth: int = 0
    #
    read_only_fields: Optional[List[str]] = Field(default=None)
    write_only_fields: Optional[Union[List[str]]] = Field(default=None)


class ModelConfig(PydanticModel):
    """
    Model Controller Configuration
    """

    allowed_routes: List[str] = Field(
        [
            "create",
            "find_one",
            "update",
            "patch",
            "delete",
            "list",
        ]
    )
    async_routes: bool = False
    create_schema: Optional[Type[PydanticModel]] = None
    retrieve_schema: Optional[Type[PydanticModel]] = None
    update_schema: Optional[Type[PydanticModel]] = None
    patch_schema: Optional[Type[PydanticModel]] = None

    pagination: Optional[ModelPagination] = Field(default=ModelPagination())
    model: Type[Model]

    schema_config: ModelSchemaConfig = Field(default=ModelSchemaConfig(exclude=set()))

    create_route_info: Dict = {}  # extra @post() information
    find_one_route_info: Dict = {}  # extra @get('/{id}') information
    update_route_info: Dict = {}  # extra @put() information
    patch_route_info: Dict = {}  # extra @patch() information
    list_route_info: Dict = {}  # extra @get('/') information
    delete_route_info: Dict = {}  # extra @delete() information

    @field_validator("allowed_routes")
    def validate_allow_routes(cls, value: List[Any]) -> Any:
        defaults = ["create", "find_one", "update", "patch", "delete", "list"]
        for item in value:
            if item not in defaults:
                raise ValueError(f"'{item}' action is not recognized in [{defaults}]")
        return value

    def __init__(self, **kwargs: Any) -> None:
        super().__init__(**kwargs)
        self.generate_all_schema()

    def generate_all_schema(self) -> None:
        _model_name = self.model.__name__.replace("Model", "")
        model_pk = getattr(
            self.model._meta.pk,
            "name",
            self.model._meta.pk.attname,  # type:ignore[union-attr]
        )

        if (
            self.create_schema
            and self.retrieve_schema
            and self.patch_schema
            and self.update_schema
        ):
            # if all schemas have been provided, then we don't need to generate any schema
            return

        if not NinjaSchemaModelSchemaConfig:  # pragma: no cover
            raise RuntimeError(
                "ninja-schema package is required for ModelControllerSchema generation.\n pip install ninja-schema"
            )

        schema_model_config = NinjaSchemaModelSchemaConfig(
            "dummy",
            ModelSchemaConfigAdapter(
                {"model": self.model, "ninja_schema_abstract": True}
            ),
        )
        all_fields = {f.name: f for f in schema_model_config.model_fields()}.keys()
        working_fields = set(all_fields)

        if self.schema_config.include == "__all__":
            working_fields = set(all_fields)

        if self.schema_config.include and self.schema_config.include != "__all__":
            include_fields = set(self.schema_config.include)
            working_fields = include_fields

        elif self.schema_config.exclude:
            exclude_fields = set(self.schema_config.exclude)
            working_fields = working_fields - exclude_fields

        if not self.create_schema and "create" in self.allowed_routes:
            create_schema_fields = self._get_create_schema_fields(
                working_fields, model_pk
            )
            self.create_schema = SchemaFactory.create_schema(
                self.model,
                name=f"{_model_name}CreateSchema",
                fields=list(create_schema_fields),
                skip_registry=True,
                depth=self.schema_config.depth,
            )

        if not self.update_schema and "update" in self.allowed_routes:
            if self.create_schema:
                self.update_schema = self.create_schema
            else:
                create_schema_fields = self._get_create_schema_fields(
                    working_fields, model_pk
                )
                self.update_schema = SchemaFactory.create_schema(
                    self.model, fields=list(create_schema_fields)
                )

        if not self.patch_schema and "patch" in self.allowed_routes:
            create_schema_fields = self._get_create_schema_fields(
                working_fields, model_pk
            )
            self.patch_schema = SchemaFactory.create_schema(
                self.model,
                name=f"{_model_name}PatchSchema",
                fields=list(create_schema_fields),
                optional_fields=list(create_schema_fields),
                skip_registry=True,
                depth=self.schema_config.depth,
            )

        if not self.retrieve_schema:
            retrieve_schema_fields = self._get_retrieve_schema_fields(
                working_fields, model_pk
            )
            self.retrieve_schema = SchemaFactory.create_schema(
                self.model,
                name=f"{_model_name}Schema",
                fields=list(retrieve_schema_fields),
                skip_registry=True,
                depth=self.schema_config.depth,
            )

    def _get_create_schema_fields(self, working_fields: set, model_pk: str) -> set:
        create_schema_fields = set(working_fields) - set(
            self.schema_config.read_only_fields or []
        )
        if self.schema_config.write_only_fields:
            invalid_key = (
                set(self.schema_config.write_only_fields) - create_schema_fields
            )
            if invalid_key:
                raise ConfigError(f"Field(s) {invalid_key} included to working fields.")
        return create_schema_fields - {model_pk}

    def _get_retrieve_schema_fields(self, working_fields: set, model_pk: str) -> set:
        retrieve_schema_fields = set(working_fields) - set(
            self.schema_config.write_only_fields or []
        )
        if self.schema_config.read_only_fields:
            invalid_key = (
                set(self.schema_config.read_only_fields) - retrieve_schema_fields
            )
            if invalid_key:
                raise ConfigError(f"Field(s) {invalid_key} included to working fields.")
        return set(list(retrieve_schema_fields) + [model_pk])



import inspect
import re
import uuid
from abc import ABC
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    cast,
    overload,
)

from django.db.models import Model, QuerySet
from django.http import HttpResponse
from django.urls import URLPattern
from django.urls import path as django_path
from injector import inject, is_decorated_with_inject
from ninja import NinjaAPI, Router
from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.security.base import AuthBase
from ninja.signature import is_async
from ninja.throttling import BaseThrottle
from ninja.utils import normalize_path

from ninja_extra.constants import ROUTE_FUNCTION, THROTTLED_FUNCTION, THROTTLED_OBJECTS
from ninja_extra.exceptions import APIException, NotFound, PermissionDenied, bad_request
from ninja_extra.helper import get_function_name
from ninja_extra.operation import Operation, PathView
from ninja_extra.permissions import AllowAny, BasePermission
from ninja_extra.permissions.base import OperationHolderMixin
from ninja_extra.shortcuts import (
    fail_silently,
    get_object_or_exception,
    get_object_or_none,
)
from ninja_extra.types import PermissionType

from .model import ModelConfig, ModelControllerBuilder, ModelService
from .registry import ControllerRegistry
from .response import Detail, Id, Ok
from .route.route_functions import AsyncRouteFunction, RouteFunction

if TYPE_CHECKING:  # pragma: no cover
    from ninja_extra import NinjaExtraAPI

    from .route.context import RouteContext

T = TypeVar("T")


class MissingAPIControllerDecoratorException(Exception):
    pass


def get_route_functions(cls: Type) -> Iterable[RouteFunction]:
    for method in cls.__dict__.values():
        if hasattr(method, ROUTE_FUNCTION):
            yield getattr(method, ROUTE_FUNCTION)


def get_all_controller_route_function(
    controller: Union[Type["ControllerBase"], Type],
) -> List[RouteFunction]:  # pragma: no cover
    route_functions: List[RouteFunction] = []
    for item in dir(controller):
        attr = getattr(controller, item)
        if isinstance(attr, RouteFunction):
            route_functions.append(attr)
    return route_functions


def compute_api_route_function(
    base_cls: Type, api_controller_instance: "APIController"
) -> None:
    for cls_route_function in get_route_functions(base_cls):
        cls_route_function.api_controller = api_controller_instance
        api_controller_instance.add_controller_route_function(cls_route_function)


class ControllerBase:
    """
    Abstract Controller Base implementation all Controller class should implement

    Example:
    ---------
    ```python
    from ninja_extra import api_controller, ControllerBase, http_get

    @api_controller
    class SomeController(ControllerBase):
        @http_get()
        def some_method_name(self):
            ...
    ```
    Inheritance Example
    -------------------
    ```python

    @api_controller
    class AnotherController(SomeController):
        @http_get()
        def some_method_name(self):
            ...
    ```
    """

    # `_api_controller` a reference to APIController instance
    _api_controller: Optional["APIController"] = None

    # `api` a reference to NinjaExtraAPI on APIController registration
    api: Optional[NinjaAPI] = None

    # `context` variable will change based on the route function called on the APIController
    # that way we can get some specific items things that belong the route function during execution
    context: Optional["RouteContext"] = None
    throttling_classes: List[Type["BaseThrottle"]] = []
    throttling_init_kwargs: Optional[Dict[Any, Any]] = None

    Ok = Ok  # TODO: remove soonest
    Id = Id  # TODO: remove soonest
    Detail = Detail  # TODO: remove soonest
    bad_request = bad_request  # TODO: remove soonest

    @classmethod
    def get_api_controller(cls) -> "APIController":
        if not cls._api_controller:
            raise MissingAPIControllerDecoratorException(
                "api_controller not found. "
                "Did you forget to use the `api_controller` decorator"
            )
        return cls._api_controller

    @classmethod
    def permission_denied(cls, permission: BasePermission) -> None:
        message = getattr(permission, "message", None)
        raise PermissionDenied(message)

    def get_object_or_exception(
        self,
        klass: Union[Type[Model], QuerySet],
        error_message: Optional[str] = None,
        exception: Type[APIException] = NotFound,
        **kwargs: Any,
    ) -> Any:
        obj = get_object_or_exception(
            klass=klass, error_message=error_message, exception=exception, **kwargs
        )
        self.check_object_permissions(obj)
        return obj

    def get_object_or_none(
        self, klass: Union[Type[Model], QuerySet], **kwargs: Any
    ) -> Optional[Any]:
        obj = get_object_or_none(klass=klass, **kwargs)
        if obj:
            self.check_object_permissions(obj)
        return obj

    def _get_permissions(self) -> Iterable[BasePermission]:
        """
        Instantiates and returns the list of permissions that this view requires.
        """
        assert self.context

        for permission_class in self.context.permission_classes:
            if isinstance(permission_class, (type, OperationHolderMixin)):
                permission_instance = permission_class()  # type: ignore[operator]
            else:
                permission_instance = permission_class
            yield permission_instance

    def check_permissions(self) -> None:
        """
        Check if the request should be permitted.
        Raises an appropriate exception if the request is not permitted.
        """
        for permission in self._get_permissions():
            if (
                self.context
                and self.context.request
                and not permission.has_permission(
                    request=self.context.request, controller=self
                )
            ):
                self.permission_denied(permission)

    def check_object_permissions(self, obj: Union[Any, Model]) -> None:
        """
        Check if the request should be permitted for a given object.
        Raises an appropriate exception if the request is not permitted.
        """
        for permission in self._get_permissions():
            if (
                self.context
                and self.context.request
                and not permission.has_object_permission(
                    request=self.context.request, controller=self, obj=obj
                )
            ):
                self.permission_denied(permission)

    def create_response(
        self, message: Any, status_code: int = 200, **kwargs: Any
    ) -> HttpResponse:
        assert self.api and self.context and self.context.request
        content = self.api.renderer.render(
            self.context.request, message, response_status=status_code
        )
        content_type = "{}; charset={}".format(
            self.api.renderer.media_type, self.api.renderer.charset
        )
        return HttpResponse(
            content, status=status_code, content_type=content_type, **kwargs
        )


class ModelControllerBase(ControllerBase):
    """
    An abstract base class for all Model Controllers

    Example:
    ---------
    ```python
    from ninja_extra import api_controller, ModelControllerBase, ModelConfig
    from .model import Post

    @api_controller
    class SomeController(ControllerBase):
        model_config = ModelConfig(model=Post)

    ```
    """

    service: ModelService
    model_config: Optional[ModelConfig] = None


ControllerClassType = TypeVar(
    "ControllerClassType",
    bound=Union[Type[ControllerBase], Type],
)


class APIController:
    _PATH_PARAMETER_COMPONENT_RE = r"{(?:(?P<converter>[^>:]+):)?(?P<parameter>[^>]+)}"

    """
    A class decorator

    Features
    --
    - Converts class to APIController
    - Forces class to inherit from `ControllerBase` if missing
    - Adapts class to Django-Ninja router

    Usage:
    ---------
    ```python
    from ninja_extra import api_controller, ControllerBase, http_post, http_get

    @api_controller
    class SomeController:
        @http_get()
        def some_method_name(self):
            ...

    assert issubclass(SomeController, ControllerBase) # true

    @api_controller
    class AnotherController(ControllerBase):
        @http_post()
        def some_method_name(self):
            ...
    ```
    You can more code intellisense within a controller decorated class when it inherits from `ControllerBase`
    as shown with `AnotherController` example.
    """

    # TODO: implement csrf on route function or on controller level. Which can override api csrf
    #   controller should have a csrf ON unless turned off by api instance

    def __init__(
        self,
        prefix: str,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        tags: Union[Optional[List[str]], str] = None,
        permissions: Optional["PermissionType"] = None,
        auto_import: bool = True,
    ) -> None:
        self.prefix = prefix
        # `auth` primarily defines APIController route function global authentication method.
        self.auth: Optional[AuthBase] = auth

        self.tags = tags  # type: ignore
        self.throttle = throttle

        self.auto_import: bool = auto_import  # set to false and it would be ignored when api.auto_discover is called
        # `controller_class` target class that the APIController wraps
        self._controller_class: Optional[Type["ControllerBase"]] = None
        # `_path_operations` a converted dict of APIController route function used by Django-Ninja library
        self._path_operations: Dict[str, PathView] = {}
        self._controller_class_route_functions: Dict[str, RouteFunction] = {}
        # `permission_classes` a collection of BasePermission Types
        # a fallback if route functions has no permissions definition
        self.permission_classes: PermissionType = permissions or [AllowAny]
        # `registered` prevents controllers from being register twice or exist in two different `api` instances
        self.registered: bool = False

        self._prefix_has_route_param = False

        if re.search(self._PATH_PARAMETER_COMPONENT_RE, prefix):
            self._prefix_has_route_param = True

        self.has_auth_async = False
        if auth and auth is not NOT_SET:
            auth_callbacks = isinstance(auth, Sequence) and auth or [auth]
            for _auth in auth_callbacks:
                _call_back = _auth if inspect.isfunction(_auth) else _auth.__call__
                if is_async(_call_back):
                    self.has_auth_async = True
                    break

    @property
    def controller_class(self) -> Type["ControllerBase"]:
        assert self._controller_class, "Controller Class is not available"
        return self._controller_class

    @property
    def tags(self) -> Optional[List[str]]:
        # `tags` is a property for grouping endpoint in Swagger API docs
        return self._tags

    @tags.setter
    def tags(self, value: Union[str, List[str], None]) -> None:
        tag: Optional[List[str]] = cast(Optional[List[str]], value)
        if tag and isinstance(value, str):
            tag = [value]
        self._tags = tag

    def __call__(self, cls: ControllerClassType) -> ControllerClassType:
        self.auto_import = getattr(cls, "auto_import", self.auto_import)
        if not issubclass(cls, ControllerBase):
            # We force the cls to inherit from `ControllerBase` by creating another type.
            cls = type(cls.__name__, (ControllerBase, cls), {"_api_controller": self})  # type:ignore[assignment]
        else:
            cls._api_controller = self

        assert isinstance(
            cls.throttling_classes, (list, tuple)
        ), f"Controller[{cls.__name__}].throttling_class must be a list or tuple"

        throttling_objects: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = (
            NOT_SET
        )

        if self.throttle is not NOT_SET:
            throttling_objects = self.throttle
        elif cls.throttling_classes:
            throttling_init_kwargs = cls.throttling_init_kwargs or {}
            throttling_objects = [
                item(**throttling_init_kwargs) for item in cls.throttling_classes
            ]

        if not self.tags:
            tag = str(cls.__name__).lower().replace("controller", "")
            self.tags = [tag]

        self._controller_class = cls

        if issubclass(cls, ModelControllerBase):
            if cls.model_config:
                # if model_config is not provided, treat controller class as normal
                builder = ModelControllerBuilder(cls, self)
                builder.register_model_routes()
                # We create a global service for handle CRUD Operations at class level
                # giving room for it to be changed at instance level through Dependency injection
                if not hasattr(cls, "service"):
                    cls.service = ModelService(cls.model_config.model)

        bases = inspect.getmro(cls)
        for base_cls in reversed(bases):
            if base_cls not in [ControllerBase, ABC, object]:
                compute_api_route_function(base_cls, self)

        for _, v in self._controller_class_route_functions.items():
            throttled_endpoint = v.as_view.__dict__.get(THROTTLED_FUNCTION)
            if v.route.route_params.throttle is NOT_SET:
                if throttled_endpoint or throttling_objects is not NOT_SET:
                    v.route.route_params.throttle = v.as_view.__dict__.get(
                        THROTTLED_OBJECTS, lambda: throttling_objects
                    )()

            self._add_operation_from_route_function(v)

        if not is_decorated_with_inject(cls.__init__):
            fail_silently(inject, constructor_or_class=cls)

        ControllerRegistry().add_controller(cls)
        return cls  # type:ignore[return-value]

    @property
    def path_operations(self) -> Dict[str, PathView]:
        return self._path_operations

    def set_api_instance(self, api: "NinjaExtraAPI") -> None:
        self.controller_class.api = api
        for path_view in self.path_operations.values():
            path_view.set_api_instance(api, cast(Router, self))

    def build_routers(self) -> List[Tuple[str, "APIController"]]:
        prefix = self.prefix
        if self._prefix_has_route_param:
            prefix = ""
        return [(prefix, self)]

    def add_controller_route_function(self, route_function: RouteFunction) -> None:
        self._controller_class_route_functions[
            get_function_name(route_function.route.view_func)
        ] = route_function

    def urls_paths(self, prefix: str) -> Iterator[URLPattern]:
        for path, path_view in self.path_operations.items():
            path = path.replace("{", "<").replace("}", ">")
            route = "/".join([i for i in (prefix, path) if i])
            # to skip lot of checks we simply treat double slash as a mistake:
            route = normalize_path(route)
            route = route.lstrip("/")
            for op in path_view.operations:
                op = cast(Operation, op)
                yield django_path(route, path_view.get_view(), name=op.url_name)

    def __repr__(self) -> str:  # pragma: no cover
        return f"<controller - {self.controller_class.__name__}>"

    def __str__(self) -> str:  # pragma: no cover
        return f"{self.controller_class.__name__}"

    def _add_operation_from_route_function(self, route_function: RouteFunction) -> None:
        # converts route functions to Operation model
        if route_function.route.route_params.operation_id is None:
            route_function.route.route_params.operation_id = f"{str(uuid.uuid4())[:8]}_controller_{route_function.route.view_func.__name__}"

        if (
            self.auth
            and self.has_auth_async
            and not isinstance(route_function, AsyncRouteFunction)
        ):
            raise Exception(
                f"You are using a Controller level Asynchronous Authentication Class, "
                f"All controller endpoint must be `async`.\n"
                f"Controller={self.controller_class.__name__}, "
                f"endpoint={get_function_name(route_function.route.view_func)}"
            )
        data = route_function.route.route_params.dict()
        route_function.operation = self.add_api_operation(
            view_func=route_function.as_view, **data
        )

    def add_api_operation(
        self,
        path: str,
        methods: List[str],
        view_func: Callable,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Operation:
        auth = self.auth if auth == NOT_SET else auth

        if self._prefix_has_route_param:
            path = normalize_path("/".join([i for i in (self.prefix, path) if i]))
        if path not in self._path_operations:
            path_view = PathView()
            self._path_operations[path] = path_view
        else:
            path_view = self._path_operations[path]
        operation = path_view.add_operation(
            path=path,
            methods=methods,
            view_func=view_func,
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
            throttle=throttle,
        )
        return operation


@overload
def api_controller(
    prefix_or_class: Union[ControllerClassType, Type[T]],
) -> Union[Type[ControllerBase], Type[T]]:  # pragma: no cover
    ...


@overload
def api_controller(
    prefix_or_class: str = "",
    auth: Any = NOT_SET,
    throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
    tags: Union[Optional[List[str]], str] = None,
    permissions: Optional["PermissionType"] = None,
    auto_import: bool = True,
) -> Callable[
    [Union[Type, Type[T]]], Union[Type[ControllerBase], Type[T]]
]:  # pragma: no cover
    ...


def api_controller(
    prefix_or_class: Union[str, Union[ControllerClassType, Type]] = "",
    auth: Any = NOT_SET,
    throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
    tags: Union[Optional[List[str]], str] = None,
    permissions: Optional["PermissionType"] = None,
    auto_import: bool = True,
) -> Union[ControllerClassType, Callable[[ControllerClassType], ControllerClassType]]:
    if isinstance(prefix_or_class, type):
        return APIController(
            prefix="",
            auth=auth,
            tags=tags,
            permissions=permissions,
            auto_import=auto_import,
            throttle=throttle,
        )(prefix_or_class)

    def _decorator(cls: ControllerClassType) -> ControllerClassType:
        return APIController(
            prefix=str(prefix_or_class),
            auth=auth,
            tags=tags,
            permissions=permissions,
            auto_import=auto_import,
            throttle=throttle,
        )(cls)

    return _decorator



import inspect
import warnings
from contextlib import contextmanager
from functools import wraps
from typing import TYPE_CHECKING, Any, Callable, Iterator, Optional, Tuple, cast

from django.http import HttpRequest, HttpResponse

from ...dependency_resolver import get_injector, service_resolver
from .context import RouteContext, get_route_execution_context

if TYPE_CHECKING:  # pragma: no cover
    from ninja_extra.operation import Operation

    from ...controllers.base import APIController, ControllerBase
    from ...controllers.route import Route


class RouteFunctionContext:
    def __init__(
        self, controller_instance: "ControllerBase", **view_func_kwargs: Any
    ) -> None:
        self.controller_instance = controller_instance
        self.view_func_kwargs = view_func_kwargs


class RouteFunction(object):
    def __init__(
        self, route: "Route", api_controller: Optional["APIController"] = None
    ):
        self.route = route
        self.operation: Optional["Operation"] = None
        self.has_request_param = False
        self.api_controller = api_controller
        self.as_view = wraps(route.view_func)(self.get_view_function())
        self._resolve_api_func_signature_(self.as_view)

    def __call__(
        self,
        request: HttpRequest,
        temporal_response: Optional[HttpResponse] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        _api_controller = self.get_api_controller()
        context = get_route_execution_context(
            request,
            temporal_response,
            self.route.permissions or _api_controller.permission_classes,  # type: ignore[arg-type]
            *args,
            **kwargs,
        )
        return self.as_view(request, *args, route_context=context, **kwargs)

    def _get_required_api_func_signature(self) -> Tuple:
        skip_parameters = ["self", "request"]
        sig_inspect = inspect.signature(self.route.view_func)
        sig_parameter = []
        for parameter in sig_inspect.parameters.values():
            if parameter.name not in skip_parameters:
                sig_parameter.append(parameter)
            elif parameter.name == "request":
                self.has_request_param = True
        return sig_inspect, sig_parameter

    def get_api_controller(self) -> "APIController":
        assert self.api_controller, "APIController is required"
        return self.api_controller

    def _resolve_api_func_signature_(self, context_func: Callable) -> Callable:
        # Override signature
        sig_inspect, sig_parameter = self._get_required_api_func_signature()
        sig_replaced = sig_inspect.replace(parameters=sig_parameter)
        context_func.__signature__ = sig_replaced  # type: ignore
        return context_func

    def get_view_function(self) -> Callable:
        def as_view(
            request: HttpRequest,
            route_context: Optional[RouteContext] = None,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            _route_context = route_context or cast(
                RouteContext, service_resolver(RouteContext)
            )
            with self._prep_controller_route_execution(_route_context, **kwargs) as ctx:
                ctx.controller_instance.check_permissions()
                result = self.route.view_func(
                    ctx.controller_instance, *args, **ctx.view_func_kwargs
                )
            return self._process_view_function_result(result)

        as_view.get_route_function = lambda: self  # type:ignore
        return as_view

    def _process_view_function_result(self, result: Any) -> Any:
        """
        This process any an returned value from view_func
        and creates an api response if result is ControllerResponseSchema
        """

        # if result and isinstance(result, ControllerResponse):
        #     return result.status_code, result.convert_to_schema()
        return result

    def _get_controller_instance(self) -> "ControllerBase":
        injector = get_injector()
        _api_controller = self.get_api_controller()

        controller_instance: "ControllerBase" = injector.create_object(
            _api_controller.controller_class
        )
        return controller_instance

    def get_route_execution_context(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> RouteContext:
        warnings.warn(
            "get_route_execution_context() is deprecated in favor of "
            "ninja_extra.controllers.route.context.get_route_execution_context()",
            DeprecationWarning,
            stacklevel=2,
        )
        _api_controller = self.get_api_controller()

        init_kwargs = {
            "permission_classes": self.route.permissions
            or _api_controller.permission_classes,
            "request": request,
            "kwargs": kwargs,
            "args": args,
        }
        context = RouteContext(**init_kwargs)  # type:ignore[arg-type]
        return context

    @contextmanager
    def _prep_controller_route_execution(
        self, route_context: RouteContext, **kwargs: Any
    ) -> Iterator[RouteFunctionContext]:
        controller_instance = self._get_controller_instance()
        controller_instance.context = route_context

        api_func_kwargs = dict(kwargs)
        if self.has_request_param:
            api_func_kwargs.update(request=route_context.request)
        try:
            yield RouteFunctionContext(
                controller_instance=controller_instance, **api_func_kwargs
            )
        except Exception as ex:
            raise ex
        finally:
            controller_instance.context = None

    def __str__(self) -> str:  # pragma: no cover
        return self.route.route_params.path

    def __repr__(self) -> str:  # pragma: no cover
        if not self.api_controller:
            return f"<RouteFunction, controller: No Controller Found, path: {self.__str__()}>"
        return f"<RouteFunction, controller: {self.api_controller.controller_class.__name__}, path: {self.__str__()}>"


class AsyncRouteFunction(RouteFunction):
    def get_view_function(self) -> Callable:
        async def as_view(
            request: HttpRequest,
            route_context: Optional[RouteContext] = None,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            from asgiref.sync import sync_to_async

            _route_context = route_context or cast(
                RouteContext, service_resolver(RouteContext)
            )
            with self._prep_controller_route_execution(_route_context, **kwargs) as ctx:
                await sync_to_async(ctx.controller_instance.check_permissions)()
                result = await self.route.view_func(
                    ctx.controller_instance, *args, **ctx.view_func_kwargs
                )
            return self._process_view_function_result(result)

        as_view.get_route_function = lambda: self  # type:ignore
        return as_view

    def __repr__(self) -> str:  # pragma: no cover
        if not self.api_controller:
            return f"<AsyncRouteFunction, controller: No Controller Found, path: {self.__str__()}>"
        return f"<AsyncRouteFunction, controller: {self.api_controller.controller_class.__name__}, path: {self.__str__()}>"

    async def __call__(
        self,
        request: HttpRequest,
        temporal_response: Optional[HttpResponse] = None,
        *args: Any,
        **kwargs: Any,
    ) -> Any:
        _api_controller = self.get_api_controller()
        context = get_route_execution_context(
            request,
            temporal_response,
            self.route.permissions or _api_controller.permission_classes,  # type:ignore[arg-type]
            *args,
            **kwargs,
        )
        return await self.as_view(request, *args, route_context=context, **kwargs)



import inspect
import operator
import re
from abc import ABC, abstractmethod
from functools import reduce, wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from asgiref.sync import sync_to_async
from django.db.models import Q, QuerySet
from django.db.models.constants import LOOKUP_SEP
from ninja import Field, Query, Schema
from ninja.constants import NOT_SET
from ninja.signature import is_async

from ninja_extra.conf import settings
from ninja_extra.shortcuts import add_ninja_contribute_args

if TYPE_CHECKING:  # pragma: no cover
    from ninja_extra.controllers import ControllerBase

__all__ = [
    "SearchingBase",
    "Searching",
    "searching",
    "SearcheratorOperation",
    "AsyncSearcheratorOperation",
]


def _istartswith(a: str, b: str) -> bool:
    return a.startswith(b)


def _isiexact(a: str, b: str) -> bool:
    return a.lower() == b.lower()


def _isiregex(a: str, b: str) -> bool:
    return bool(re.search(b, a, re.IGNORECASE))


def _isicontains(a: str, b: str) -> bool:
    return b.lower() in a.lower()


class SearchingBase(ABC):
    class Input(Schema): ...

    InputSource = Query(...)

    def __init__(self, *, pass_parameter: Optional[str] = None, **kwargs: Any) -> None:
        self.pass_parameter = pass_parameter

    @abstractmethod
    def searching_queryset(
        self, items: Union[QuerySet, List], searching_input: Any
    ) -> Union[QuerySet, List]: ...


class Searching(SearchingBase):
    class Input(Schema):
        search: Optional[str] = Field(None)

    lookup_prefixes = {
        "^": "istartswith",
        "=": "iexact",
        "@": "search",
        "$": "iregex",
    }

    lookup_prefixes_list = {
        "^": _istartswith,
        "=": _isiexact,
        "$": _isiregex,
    }

    def __init__(
        self,
        search_fields: Optional[List[str]] = None,
        pass_parameter: Optional[str] = None,
    ) -> None:
        super().__init__(pass_parameter=pass_parameter)
        self.search_fields = search_fields or []

    def searching_queryset(
        self, items: Union[QuerySet, List], searching_input: Input
    ) -> Union[QuerySet, List]:
        search_terms = self.get_search_terms(searching_input.search)

        if self.search_fields and search_terms:
            if isinstance(items, QuerySet):
                conditions_queryset = self.construct_conditions_for_queryset(
                    search_terms
                )
                return items.filter(reduce(operator.and_, conditions_queryset))
            elif isinstance(items, list):
                conditions_list = self.construct_conditions_for_list(search_terms)
                return [
                    item for item in items if self.filter_spec(item, conditions_list)
                ]
        return items

    def get_search_terms(self, value: Optional[str]) -> List[str]:
        if value:
            value = value.replace("\x00", "")  # strip null characters
            value = value.replace(",", " ")
            return value.split()
        return []

    def construct_search(self, field_name: str) -> str:
        lookup = self.lookup_prefixes.get(field_name[0])
        if lookup:
            field_name = field_name[1:]
        else:
            lookup = "icontains"
        return LOOKUP_SEP.join([field_name, lookup])

    def construct_conditions_for_queryset(self, search_terms: List[str]) -> List[Q]:
        orm_lookups = [
            self.construct_search(str(search_field))
            for search_field in self.search_fields
        ]
        conditions = []
        for search_term in search_terms:
            queries = [Q(**{orm_lookup: search_term}) for orm_lookup in orm_lookups]
            conditions.append(reduce(operator.or_, queries))
        return conditions

    def construct_conditions_for_list(
        self, search_terms: List[str]
    ) -> Dict[str, List[Tuple[Callable, str]]]:
        lookups = self.construct_search_for_list()
        conditions: Dict[str, List[Tuple[Callable, str]]] = {
            field_name: [] for field_name in lookups.keys()
        }
        for search_term in search_terms:
            for field_name, lookup in lookups.items():
                conditions[field_name].append((lookup, search_term))
        return conditions

    def construct_search_for_list(self) -> Dict[str, Callable]:
        def get_lookup(prefix: str) -> Callable:
            return self.lookup_prefixes_list.get(prefix, _isicontains)

        return {
            field_name[1:]
            if (self.lookup_prefixes_list.get(field_name[0]))
            else field_name: get_lookup(field_name[0])
            for field_name in self.search_fields
        }

    def filter_spec(
        self, item: Any, conditions: Dict[str, List[Tuple[Callable, str]]]
    ) -> bool:
        item_getter = (
            operator.itemgetter if isinstance(item, dict) else operator.attrgetter
        )
        for field, lookup in conditions.items():
            if not any(
                lookup_func(item_getter(field)(item), lookup_value)
                for lookup_func, lookup_value in lookup
            ):
                return False
        return True


@overload
def searching() -> Callable[..., Any]:  # pragma: no cover
    ...


@overload
def searching(
    func_or_searching_class: Any = NOT_SET, **searching_params: Any
) -> Callable[..., Any]:  # pragma: no cover
    ...


def searching(
    func_or_searching_class: Any = NOT_SET, **searching_params: Any
) -> Callable[..., Any]:
    isfunction = inspect.isfunction(func_or_searching_class)
    isnotset = func_or_searching_class == NOT_SET

    searching_class: Type[SearchingBase] = settings.SEARCHING_CLASS

    if isfunction:
        return _inject_searcherator(func_or_searching_class, searching_class)

    if not isnotset:
        searching_class = func_or_searching_class

    def wrapper(func: Callable[..., Any]) -> Any:
        return _inject_searcherator(func, searching_class, **searching_params)

    return wrapper


def _inject_searcherator(
    func: Callable[..., Any],
    searching_class: Type[SearchingBase],
    **searching_params: Any,
) -> Callable[..., Any]:
    searcherator: SearchingBase = searching_class(**searching_params)
    searcherator_kwargs_name = "searching"
    searcherator_operation_class = SearcheratorOperation
    if is_async(func):
        searcherator_operation_class = AsyncSearcheratorOperation
    searcherator_operation = searcherator_operation_class(
        searcherator=searcherator,
        view_func=func,
        searcherator_kwargs_name=searcherator_kwargs_name,
    )

    return searcherator_operation.as_view


class SearcheratorOperation:
    def __init__(
        self,
        *,
        searcherator: SearchingBase,
        view_func: Callable,
        searcherator_kwargs_name: str = "searching",
    ) -> None:
        self.searcherator = searcherator
        self.searcherator_kwargs_name = searcherator_kwargs_name
        self.view_func = view_func

        searcherator_view = self.get_view_function()
        self.as_view = wraps(view_func)(searcherator_view)
        add_ninja_contribute_args(
            self.as_view,
            (
                self.searcherator_kwargs_name,
                self.searcherator.Input,
                self.searcherator.InputSource,
            ),
        )
        searcherator_view.searcherator_operation = self  # type:ignore[attr-defined]

    @property
    def view_func_has_kwargs(self) -> bool:  # pragma: no cover
        return self.searcherator.pass_parameter is not None

    def get_view_function(self) -> Callable:
        def as_view(controller: "ControllerBase", *args: Any, **kw: Any) -> Any:
            func_kwargs = dict(**kw)
            searching_params = func_kwargs.pop(self.searcherator_kwargs_name)
            if self.searcherator.pass_parameter:
                func_kwargs[self.searcherator.pass_parameter] = searching_params

            items = self.view_func(controller, *args, **func_kwargs)
            return self.searcherator.searching_queryset(items, searching_params)

        return as_view


class AsyncSearcheratorOperation(SearcheratorOperation):
    def get_view_function(self) -> Callable:
        async def as_view(controller: "ControllerBase", *args: Any, **kw: Any) -> Any:
            func_kwargs = dict(**kw)
            searching_params = func_kwargs.pop(self.searcherator_kwargs_name)
            if self.searcherator.pass_parameter:
                func_kwargs[self.searcherator.pass_parameter] = searching_params

            items = await self.view_func(controller, *args, **func_kwargs)
            searching_queryset = cast(
                Callable, sync_to_async(self.searcherator.searching_queryset)
            )
            return await searching_queryset(items, searching_params)

        return as_view



import inspect
import time
from contextlib import contextmanager
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Sequence,
    Type,
    Union,
    cast,
)

from django.http import HttpRequest
from django.http.response import HttpResponse, HttpResponseBase
from django.utils.encoding import force_str
from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.errors import AuthenticationError
from ninja.operation import (
    AsyncOperation as NinjaAsyncOperation,
)
from ninja.operation import (
    Operation as NinjaOperation,
)
from ninja.operation import (
    PathView as NinjaPathView,
)
from ninja.signature import is_async
from ninja.throttling import BaseThrottle
from ninja.types import TCallable
from ninja.utils import check_csrf

from ninja_extra.compatible import asynccontextmanager
from ninja_extra.constants import ROUTE_CONTEXT_VAR
from ninja_extra.exceptions import APIException, Throttled
from ninja_extra.helper import get_function_name
from ninja_extra.logger import request_logger

# from ninja_extra.signals import route_context_finished, route_context_started
from ninja_extra.types import PermissionType

from .controllers.route.context import RouteContext, get_route_execution_context
from .details import ViewSignature

if TYPE_CHECKING:  # pragma: no cover
    from .controllers.route.route_functions import RouteFunction


class Operation(NinjaOperation):
    def __init__(
        self,
        path: str,
        methods: List[str],
        view_func: Callable,
        *,
        url_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        self.is_coroutine = is_async(view_func)
        self.url_name = url_name  # type: ignore[assignment]
        super().__init__(path, methods, view_func, **kwargs)
        self.signature = ViewSignature(self.path, self.view_func)

    def _set_auth(
        self, auth: Optional[Union[Sequence[Callable], Callable, object]]
    ) -> None:
        if auth is not None and auth is not NOT_SET:
            self.auth_callbacks = isinstance(auth, Sequence) and auth or [auth]
            for callback in self.auth_callbacks:
                _call_back = (
                    callback if inspect.isfunction(callback) else callback.__call__  # type: ignore
                )

                if not getattr(callback, "is_coroutine", None):
                    callback.is_coroutine = is_async(  # type:ignore[union-attr]
                        _call_back
                    )

                if is_async(_call_back) and not self.is_coroutine:
                    raise Exception(
                        f"Could apply auth=`{get_function_name(callback)}` "
                        f"to view_func=`{get_function_name(self.view_func)}`.\n"
                        f"N:B - {get_function_name(callback)} can only be used on Asynchronous view functions"
                    )

    def _log_action(
        self,
        logger: Callable[..., Any],
        request: HttpRequest,
        duration: Optional[float] = None,
        ex: Optional[Exception] = None,
        **kwargs: Any,
    ) -> None:
        try:
            msg = (
                f'"{request.method.upper() if request.method else "METHOD NOT FOUND"} - '
                f'{self.view_func.__name__} {request.path}" '
                f"{duration if duration else ''}"
            )
            if hasattr(self.view_func, "get_route_function"):
                route_function: "RouteFunction" = self.view_func.get_route_function()
                api_controller = route_function.get_api_controller()

                msg = (
                    f'"{request.method.upper() if request.method else "METHOD NOT FOUND"} - '
                    f'{api_controller.controller_class.__name__}[{self.view_func.__name__}] {request.path}" '
                    f"{duration if duration else ''}"
                )
            if ex:
                msg += (
                    f"{ex.status_code}"
                    if isinstance(ex, APIException)
                    else f"{force_str(ex.args)}"
                )

            logger(msg, **kwargs)
        except Exception as log_ex:
            request_logger.debug(log_ex)

    def get_execution_context(
        self,
        request: HttpRequest,
        temporal_response: HttpResponse,
        *args: Any,
        **kwargs: Any,
    ) -> RouteContext:
        permission_classes: PermissionType = []
        if hasattr(self.view_func, "get_route_function"):
            route_function: "RouteFunction" = self.view_func.get_route_function()

            _api_controller = route_function.get_api_controller()
            permission_classes = (
                route_function.route.permissions or _api_controller.permission_classes  # type: ignore[assignment]
            )

        return get_route_execution_context(
            request,
            temporal_response,
            permission_classes,
            *args,
            **kwargs,
        )

    @contextmanager
    def _prep_run(
        self, request: HttpRequest, temporal_response: HttpResponse, **kw: Any
    ) -> Iterator[RouteContext]:
        try:
            start_time = time.time()
            context = self.get_execution_context(
                request, temporal_response=temporal_response, **kw
            )
            # send route_context_started signal
            ROUTE_CONTEXT_VAR.set(context)

            yield context
            self._log_action(
                request_logger.info,
                request=request,
                duration=time.time() - start_time,
                extra={"request": request},
                exc_info=None,
            )
        except Exception as e:
            self._log_action(
                request_logger.warning,
                request=request,
                ex=e,
                extra={"request": request},
                exc_info=None,
            )
            raise e
        finally:
            # send route_context_finished signal
            ROUTE_CONTEXT_VAR.set(None)

    def run(self, request: HttpRequest, **kw: Any) -> HttpResponseBase:
        try:
            temporal_response = self.api.create_temporal_response(request)
            with self._prep_run(
                request, temporal_response=temporal_response, **kw
            ) as ctx:
                error = self._run_checks(request)
                if error:
                    return error

                values = self._get_values(request, kw, temporal_response)
                ctx.kwargs.update(values)
                result = self.view_func(request, **values)
                _processed_results = self._result_to_response(
                    request, result, temporal_response
                )
                return _processed_results
        except Exception as e:
            if isinstance(e, TypeError) and "required positional argument" in str(
                e
            ):  # pragma: no cover
                msg = "Did you fail to use functools.wraps() in a decorator?"
                msg = f"{e.args[0]}: {msg}" if e.args else msg
                e.args = (msg,) + e.args[1:]
            return self.api.on_exception(request, e)

    def _check_throttles(self, request: HttpRequest) -> Optional[HttpResponse]:
        throttle_durations = []
        for throttle in self.throttle_objects:
            if not throttle.allow_request(request):
                throttle_durations.append(throttle.wait())

        if throttle_durations:
            # Filter out `None` values which may happen in case of config / rate
            durations = [
                duration for duration in throttle_durations if duration is not None
            ]

            duration = max(durations, default=None)
            raise Throttled(wait=duration)
        return None


class AsyncOperation(Operation, NinjaAsyncOperation):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        from asgiref.sync import sync_to_async

        self._get_values = cast(Callable, sync_to_async(super()._get_values))  # type: ignore
        self._result_to_response = cast(  # type: ignore
            Callable,
            sync_to_async(super()._result_to_response),
        )

    async def _run_checks(self, request: HttpRequest) -> Optional[HttpResponse]:  # type: ignore
        """Runs security checks for each operation"""
        # csrf:
        if self.api.csrf:
            error = check_csrf(request, self.view_func)
            if error:
                return error

        # auth:
        if self.auth_callbacks:
            error = await self._run_authentication(request)  # type: ignore[assignment]
            if error:
                return error

        # Throttling:
        if self.throttle_objects:
            error = self._check_throttles(request)  # type: ignore
            if error:
                return error

        return None

    async def _run_authentication(self, request: HttpRequest) -> Optional[HttpResponse]:  # type: ignore
        for callback in self.auth_callbacks:
            try:
                is_coroutine = getattr(callback, "is_coroutine", False)
                if is_coroutine:
                    result = await callback(request)
                else:
                    result = callback(request)
            except Exception as exc:
                return self.api.on_exception(request, exc)

            if result:
                request.auth = result  # type: ignore
                return None

        return self.api.on_exception(request, AuthenticationError())

    @asynccontextmanager
    async def _prep_run(  # type:ignore
        self, request: HttpRequest, **kw: Any
    ) -> AsyncIterator[RouteContext]:
        try:
            start_time = time.time()
            context = self.get_execution_context(request, **kw)
            # send route_context_started signal
            ROUTE_CONTEXT_VAR.set(context)

            yield context
            self._log_action(
                request_logger.info,
                request=request,
                duration=time.time() - start_time,
                extra={"request": request},
                exc_info=None,
            )
        except Exception as e:
            self._log_action(
                request_logger.warning,
                request=request,
                ex=e,
                extra={"request": request},
                exc_info=None,
            )
            raise e
        finally:
            # send route_context_finished signal
            ROUTE_CONTEXT_VAR.set(None)

    async def run(self, request: HttpRequest, **kw: Any) -> HttpResponseBase:  # type: ignore
        try:
            temporal_response = self.api.create_temporal_response(request)
            async with self._prep_run(
                request, temporal_response=temporal_response, **kw
            ) as ctx:
                error = await self._run_checks(request)
                if error:
                    return error

                values = await self._get_values(request, kw, temporal_response)  # type: ignore
                ctx.kwargs.update(values)
                result = await self.view_func(request, **values)
                _processed_results = await self._result_to_response(
                    request, result, temporal_response
                )
                return cast(HttpResponseBase, _processed_results)
        except Exception as e:
            return self.api.on_exception(request, e)


class PathView(NinjaPathView):
    async def _async_view(  # type: ignore
        self, request: HttpRequest, *args, **kwargs
    ) -> HttpResponseBase:
        return await super(PathView, self)._async_view(request, *args, **kwargs)

    def _sync_view(self, request: HttpRequest, *args, **kwargs) -> HttpResponseBase:  # type: ignore
        return super(PathView, self)._sync_view(request, *args, **kwargs)

    def add_operation(
        self,
        path: str,
        methods: List[str],
        view_func: Callable,
        *,
        auth: Optional[Union[Sequence[Callable], Callable, object]] = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> Operation:
        if url_name:
            self.url_name = url_name
        operation_class = self.get_operation_class(view_func)
        operation = operation_class(
            path,
            methods,
            view_func,
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            url_name=url_name,
            openapi_extra=openapi_extra,
            throttle=throttle,
        )

        self.operations.append(operation)
        return operation

    def get_operation_class(
        self, view_func: TCallable
    ) -> Type[Union[Operation, AsyncOperation]]:
        operation_class = Operation
        if is_async(view_func):
            self.is_async = True
            operation_class = AsyncOperation
        return operation_class



import inspect
import warnings
from typing import TYPE_CHECKING

from ninja.signature.details import ViewSignature as NinjaViewSignature

if TYPE_CHECKING:  # pragma: no cover
    from .controllers.route.route_functions import RouteFunction


class ViewSignature(NinjaViewSignature):
    def _validate_view_path_params(self) -> None:
        """verify all path params are present in the path model fields"""
        if self.path_params_names:
            path_model = next(
                (m for m in self.models if m.__ninja_param_source__ == "path"), None
            )
            missing = tuple(
                sorted(
                    name
                    for name in self.path_params_names
                    if not (path_model and name in path_model.__ninja_flatten_map__)
                )
            )
            if missing:  # pragma: no cover
                message = f"Field(s) {missing} are in the view path, but were not found in the view signature."
                view_func = self.view_func
                filename = inspect.getfile(view_func)

                if hasattr(self.view_func, "get_route_function"):
                    route_function: "RouteFunction" = (
                        self.view_func.get_route_function()
                    )
                    api_controller = route_function.get_api_controller()

                    view_func = route_function.route.view_func

                    message += f" @ {api_controller.controller_class.__name__}[{self.view_func.__name__}]"
                    filename = inspect.getfile(api_controller.controller_class)

                warnings.warn_explicit(
                    UserWarning(message),
                    category=None,
                    filename=filename,
                    lineno=inspect.getsourcelines(view_func)[1],
                    source=None,
                )



import inspect
import logging
from typing import (
    Any,
    Callable,
    Type,
    overload,
)

from ninja.constants import NOT_SET
from ninja.pagination import PaginationBase
from ninja.signature import is_async

from ninja_extra.conf import settings
from ninja_extra.pagination.operations import (
    AsyncPaginatorOperation,
    PaginatorOperation,
)

logger = logging.getLogger()


@overload
def paginate() -> Callable[..., Any]:  # pragma: no cover
    ...


@overload
def paginate(
    func_or_pgn_class: Any = NOT_SET, **paginator_params: Any
) -> Callable[..., Any]:  # pragma: no cover
    ...


def paginate(
    func_or_pgn_class: Any = NOT_SET, **paginator_params: Any
) -> Callable[..., Any]:
    isfunction = inspect.isfunction(func_or_pgn_class)
    is_not_set = func_or_pgn_class == NOT_SET

    pagination_class: Type[PaginationBase] = settings.PAGINATION_CLASS

    if isfunction:
        return _inject_pagination(func_or_pgn_class, pagination_class)

    if not is_not_set:
        pagination_class = func_or_pgn_class

    def wrapper(func: Callable[..., Any]) -> Any:
        return _inject_pagination(func, pagination_class, **paginator_params)

    return wrapper


def _inject_pagination(
    func: Callable[..., Any],
    paginator_class: Type[PaginationBase],
    **paginator_params: Any,
) -> Callable[..., Any]:
    paginator: PaginationBase = paginator_class(**paginator_params)
    paginator_kwargs_name = "pagination"
    paginator_operation_class = PaginatorOperation

    if is_async(func):
        paginator_operation_class = AsyncPaginatorOperation
    paginator_operation = paginator_operation_class(
        paginator=paginator, view_func=func, paginator_kwargs_name=paginator_kwargs_name
    )

    return paginator_operation.as_view



import inspect
from typing import Any, Callable, List, Type, Union

from ninja.throttling import BaseThrottle

from ninja_extra.conf import settings
from ninja_extra.constants import THROTTLED_FUNCTION, THROTTLED_OBJECTS


def throttle(
    *func_or_throttle_klass_or_object: Any, **init_kwargs: Any
) -> Callable[..., Any]:
    isfunction = (
        inspect.isfunction(func_or_throttle_klass_or_object[0])
        if len(func_or_throttle_klass_or_object) == 1
        else False
    )

    if isfunction:
        func = func_or_throttle_klass_or_object[0]
        throttle_classes: List[Type[BaseThrottle]] = settings.THROTTLE_CLASSES
        return _inject_throttling(func, *throttle_classes, **init_kwargs)

    def wrapper(view_func: Callable[..., Any]) -> Any:
        return _inject_throttling(
            view_func, *func_or_throttle_klass_or_object, **init_kwargs
        )

    return wrapper


# def _run_throttles(
#     *throttle_classes: Type[BaseThrottle],
#     request_or_controller: Union[HttpRequest, ControllerBase],
#     response: Optional[HttpResponse] = None,
#     **init_kwargs: Any,
# ) -> None:
#     """
#     Run all throttles for a request.
#     Raises an appropriate exception if the request is throttled.
#     """
#
#     request = cast(
#         HttpRequest,
#         (
#             request_or_controller.context.request  # type:ignore
#             if isinstance(request_or_controller, ControllerBase)
#             else request_or_controller
#         ),
#     )
#
#     throttle_durations = []
#
#     for throttle_class in throttle_classes:
#         throttling: BaseThrottle = throttle_class(**init_kwargs)
#         if not throttling.allow_request(request):
#             # Filter out `None` values which may happen in case of config / rate
#             duration = throttling.wait()
#             if duration is not None:
#                 throttle_durations.append(duration)
#
#     if throttle_durations:
#         duration = max(throttle_durations, default=None)
#         raise exceptions.Throttled(duration)


def _lazy_throttling_objects(
    *throttle_klass_or_object: Union[Type[BaseThrottle], BaseThrottle],
    **init_kwargs: Any,
) -> Callable:
    def _() -> List[BaseThrottle]:
        res = []
        for item in throttle_klass_or_object:
            if isinstance(item, type):
                res.append(item(**init_kwargs))
                continue

            res.append(item)

        return res

    return _


def _inject_throttling(
    func: Callable[..., Any],
    *throttle_classes: Type[BaseThrottle],
    **init_kwargs: Any,
) -> Callable[..., Any]:
    throttling_objects_lazy: Callable = _lazy_throttling_objects(
        *throttle_classes, **init_kwargs
    )
    setattr(func, THROTTLED_FUNCTION, True)
    setattr(func, THROTTLED_OBJECTS, throttling_objects_lazy)

    return func
    # if is_async(func):
    #     return _async_inject_throttling_handler(func, *throttle_classes, **init_kwargs)
    # return _sync_inject_throttling_handler(func, *throttle_classes, **init_kwargs)


# def _sync_inject_throttling_handler(
#     func: Callable[..., Any],
#     *throttle_classes: Type[BaseThrottle],
#     **init_kwargs: Any,
# ) -> Callable[..., Any]:
#     @wraps(func)
#     def as_view(
#         request_or_controller: Union[HttpRequest, ControllerBase], *args: Any, **kw: Any
#     ) -> Any:
#         ctx = cast(Optional[RouteContext], service_resolver(RouteContext))
#         _run_throttles(
#             *throttle_classes,
#             request_or_controller=request_or_controller,
#             response=ctx.response if ctx else None,
#             **init_kwargs,
#         )
#
#         res = func(request_or_controller, *args, **kw)
#         return res
#
#     return as_view
#
#
# def _async_inject_throttling_handler(
#     func: Callable[..., Any],
#     *throttle_classes: Type[BaseThrottle],
#     **init_kwargs: Any,
# ) -> Callable[..., Any]:
#     @wraps(func)
#     async def as_view(
#         request_or_controller: Union[HttpRequest, ControllerBase], *args: Any, **kw: Any
#     ) -> Any:
#         ctx = cast(Optional[RouteContext], service_resolver(RouteContext))
#         _run_throttles(
#             *throttle_classes,
#             request_or_controller=request_or_controller,
#             response=ctx.response if ctx else None,
#             **init_kwargs,
#         )
#
#         res = await func(request_or_controller, *args, **kw)
#         return res
#
#     return as_view



from typing import TYPE_CHECKING, Dict, Optional, Type

if TYPE_CHECKING:  # pragma: no cover
    from ninja_extra.controllers.base import ControllerBase  # pragma: no cover


class ControllerBorg:
    _shared_state_: Dict[str, Dict[str, Type["ControllerBase"]]] = {"controllers": {}}

    def __init__(self) -> None:
        self.__dict__ = self._shared_state_

    def add_controller(self, controller: Type["ControllerBase"]) -> None:
        if (
            hasattr(controller, "get_api_controller")
            and controller.get_api_controller().auto_import
        ):
            self._shared_state_["controllers"].update({str(controller): controller})

    def remove_controller(
        self, controller: Type["ControllerBase"]
    ) -> Optional[Type["ControllerBase"]]:
        if str(controller) in self._shared_state_["controllers"]:
            return self._shared_state_["controllers"].pop(str(controller))
        return None

    def clear_controller(self) -> None:
        self._shared_state_["controllers"] = {}

    @classmethod
    def get_controllers(cls) -> Dict[str, Type["ControllerBase"]]:
        return cls._shared_state_.get("controllers", {})


class ControllerRegistry(ControllerBorg):
    def __init__(self) -> None:
        ControllerBorg.__init__(self)



import typing as t

from .response import (
    NinjaPaginationResponseSchema,
    PaginatedResponseSchema,
    RouteParameter,
)

__all__ = ["PaginatedResponseSchema", "RouteParameter", "NinjaPaginationResponseSchema"]


def __getattr__(name: str) -> t.Any:  # pragma: no cover
    if name in [
        "IdSchema",
        "OkSchema",
        "DetailSchema",
    ]:
        raise RuntimeError(f"'{name}' is no longer available")



from typing import Any, Callable, Dict, List, Optional, Union, get_type_hints

from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.router import Router as NinjaRouter
from ninja.throttling import BaseThrottle

from ninja_extra.operation import PathView

__all__ = ["Router"]


class Router(NinjaRouter):
    def __init__(
        self,
        *,
        auth: Any = NOT_SET,
        tags: Optional[List[str]] = None,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
    ) -> None:
        super().__init__(auth=auth, tags=tags, throttle=throttle)
        self.path_operations: Dict[str, PathView] = {}  # type: ignore

    def add_api_operation(
        self,
        path: str,
        methods: List[str],
        view_func: Callable,
        *,
        auth: Any = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: Any = NOT_SET,
        operation_id: Optional[str] = None,
        summary: Optional[str] = None,
        description: Optional[str] = None,
        tags: Optional[List[str]] = None,
        deprecated: Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: Optional[str] = None,
        include_in_schema: bool = True,
        openapi_extra: Optional[Dict[str, Any]] = None,
    ) -> None:
        if path not in self.path_operations:
            path_view = PathView()
            self.path_operations[path] = path_view
        else:
            path_view = self.path_operations[path]

        if response is NOT_SET:
            response = get_type_hints(view_func).get("return") or NOT_SET

        path_view.add_operation(
            path=path,
            throttle=throttle,
            methods=methods,
            view_func=view_func,
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
        )
        if self.api:
            path_view.set_api_instance(self.api, self)

        return None



from typing import Any, Dict, Optional, Tuple, Type, cast, no_type_check

_generic_types_registry: Dict[Tuple[Type["GenericType"], Type], Type] = {}


class GenericModelMeta(type):
    def _get_name(self: Type["GenericType"], wrap_type: Type) -> str:  # type: ignore
        _name = f"{self._generic_base_name or self.__name__}"
        if hasattr(wrap_type, "__name__"):
            _name += f"[{str(wrap_type.__name__)}]"
        return _name

    def __getitem__(self: Type["GenericType"], wraps: Any) -> Any:  # type: ignore
        if (self, wraps) not in _generic_types_registry:
            new_generic_type = cast(Type, self().get_generic_type(wraps))
            new_generic_type.__generic_model__ = self
            new_generic_type.__name__ = self._get_name(wraps)

            _generic_types_registry[self, wraps] = new_generic_type
        return _generic_types_registry[self, wraps]


@no_type_check
class GenericType(metaclass=GenericModelMeta):
    _generic_base_name: Optional[str] = None
    """
    Get a wrapper class that mimic python 3.8 generic support to python3.6, 3.7
    Examples
    --------
    Create a Class to reference the generic model to be create:
    >>> class ObjectA(GenericType, generic_base_name="A"):
    >>>     def get_generic_type(self, wrap_type):
    >>>         class ObjectAGeneric(self):
    >>>             item: wrap_type
    >>>         return ObjectAGeneric

    Usage:
    >>> class Pass: ...
    >>> object_generic_type  = ObjectA[Pass]
    """

    def __init_subclass__(cls, **kwargs: Any) -> None:
        cls._generic_base_name = kwargs.get("generic_base_name")
        return super().__init_subclass__()

    def get_generic_type(self, wrap_type: Any) -> Any:  # pragma: no cover
        raise NotImplementedError



from .builder import ModelControllerBuilder
from .endpoints import ModelAsyncEndpointFactory, ModelEndpointFactory
from .interfaces import ModelServiceBase
from .schemas import ModelConfig, ModelPagination, ModelSchemaConfig
from .service import ModelService

__all__ = [
    "ModelServiceBase",
    "ModelService",
    "ModelConfig",
    "ModelSchemaConfig",
    "ModelPagination",
    "ModelControllerBuilder",
    "ModelEndpointFactory",
    "ModelAsyncEndpointFactory",
]



import typing as t
import uuid

from django.db.models import Model as DjangoModel
from django.db.models import QuerySet
from ninja.pagination import PaginationBase
from ninja.params import Body
from ninja.signature import is_async
from pydantic import BaseModel as PydanticModel

from ninja_extra.permissions import BasePermission

from ... import status
from ...exceptions import NotFound
from ...pagination import PageNumberPaginationExtra, PaginatedResponseSchema, paginate
from ..route import route
from .path_resolver import AsyncPathResolverOperation, PathResolverOperation

if t.TYPE_CHECKING:
    from ..base import ModelControllerBase


async def check_if_coroutine(func_result: t.Union[t.Any, t.Coroutine]) -> t.Any:
    if isinstance(func_result, t.Coroutine):
        return await func_result
    return func_result


def path_resolver(path: str, func: t.Callable) -> t.Callable:
    resolver_class = PathResolverOperation
    if is_async(func):
        resolver_class = AsyncPathResolverOperation

    instance = resolver_class(path, func)
    return instance.as_view


class ModelEndpointFactory:
    """
    Factory for creating CRUD operations of a model controller and for adding custom route functions to controllers.

    example:
    ```python

    api_controller
    class SampleMocelController(ModelControllerBase):

        create_sample = ModelEndpointFactory.create()
        update_sample = ModelEndpointFactory.update()

        delete_sample = ModelEndpointFactory.delete()
        patch_sample = ModelEndpointFactory.patch()

        get_sample = ModelEndpointFactory.get()
        list_samples = ModelEndpointFactory.list()
    ```
    """

    @classmethod
    def _clean_path(cls, path: str) -> str:
        working_path = path.split("?")
        return working_path[0]

    @classmethod
    def _change_name(cls, name_prefix: str) -> str:
        unique = str(uuid.uuid4())[:6]
        return f"{name_prefix}_{unique}"

    @classmethod
    def create(
        cls,
        schema_in: t.Type[PydanticModel],
        schema_out: t.Type[PydanticModel],
        path: str = "/",
        status_code: int = status.HTTP_201_CREATED,
        url_name: t.Optional[str] = None,
        custom_handler: t.Optional[t.Callable[..., t.Any]] = None,
        description: t.Optional[str] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Create new item",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a POST Action
        """
        working_path = cls._clean_path(path)

        create_item = path_resolver(
            path,
            cls._create_handler(schema_in=schema_in, custom_handler=custom_handler),
        )
        return route.post(
            working_path,
            response={status_code: schema_out},
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )(create_item)

    @classmethod
    def update(
        cls,
        path: str,
        lookup_param: str,
        schema_in: t.Type[PydanticModel],
        schema_out: t.Type[PydanticModel],
        status_code: int = status.HTTP_200_OK,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        object_getter: t.Optional[t.Callable[..., DjangoModel]] = None,
        custom_handler: t.Optional[t.Callable[..., t.Any]] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Update an item",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a PUT Action
        """
        working_path = cls._clean_path(path)
        update_item = path_resolver(
            path,
            cls._update_handler(
                schema_in=schema_in,
                object_getter=object_getter,
                lookup_param=lookup_param,
                custom_handler=custom_handler,
            ),
        )
        return route.put(
            working_path,
            response={status_code: schema_out},
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )(update_item)

    @classmethod
    def patch(
        cls,
        path: str,
        lookup_param: str,
        schema_in: t.Type[PydanticModel],
        schema_out: t.Type[PydanticModel],
        status_code: int = status.HTTP_200_OK,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        object_getter: t.Optional[t.Callable[..., DjangoModel]] = None,
        custom_handler: t.Optional[t.Callable[..., t.Any]] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Patch Item Update",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a PATCH Action
        """
        working_path = cls._clean_path(path)
        patch_item = path_resolver(
            path,
            cls._patch_handler(
                object_getter=object_getter,
                custom_handler=custom_handler,
                lookup_param=lookup_param,
                schema_in=schema_in,
            ),
        )
        return route.patch(
            working_path,
            response={status_code: schema_out},
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )(patch_item)

    @classmethod
    def find_one(
        cls,
        path: str,
        lookup_param: str,
        schema_out: t.Type[PydanticModel],
        status_code: int = status.HTTP_200_OK,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        object_getter: t.Optional[t.Callable[..., DjangoModel]] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Find a specific item",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a GET Action
        """
        working_path = cls._clean_path(path)
        get_item = path_resolver(
            path,
            cls._find_one_handler(
                object_getter=object_getter, lookup_param=lookup_param
            ),
        )
        return route.get(
            working_path,
            response={status_code: schema_out},
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )(get_item)

    @classmethod
    def list(
        cls,
        schema_out: t.Type[PydanticModel],
        path: str = "/",
        status_code: int = status.HTTP_200_OK,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "List of Items",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
        queryset_getter: t.Optional[t.Callable[..., QuerySet]] = None,
        pagination_response_schema: t.Optional[
            t.Type[PydanticModel]
        ] = PaginatedResponseSchema,
        pagination_class: t.Optional[
            t.Type[PaginationBase]
        ] = PageNumberPaginationExtra,
        **paginate_kwargs: t.Any,
    ) -> t.Callable:
        """
        Creates a GET Action to list Items
        """
        working_path = cls._clean_path(path)
        list_items = path_resolver(
            path, cls._list_handler(queryset_getter=queryset_getter)
        )

        if pagination_response_schema and pagination_class:
            list_items = paginate(pagination_class, **paginate_kwargs)(list_items)
            return route.get(
                working_path,
                response={
                    status_code: pagination_response_schema[schema_out]  # type:ignore[index]
                },
                url_name=url_name,
                description=description,
                operation_id=operation_id,
                summary=summary,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
            )(list_items)

        return route.get(
            working_path,
            response={status_code: t.List[schema_out]},  # type:ignore[valid-type]
            url_name=url_name,
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )(list_items)

    @classmethod
    def delete(
        cls,
        path: str,
        lookup_param: str,
        status_code: int = status.HTTP_204_NO_CONTENT,
        url_name: t.Optional[str] = None,
        description: t.Optional[str] = None,
        object_getter: t.Optional[t.Callable[..., DjangoModel]] = None,
        custom_handler: t.Optional[t.Callable[..., t.Any]] = None,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = "Delete An Item",
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable:
        """
        Creates a DELETE Action to list Items
        """
        working_path = cls._clean_path(path)
        delete_item = path_resolver(
            path,
            cls._delete_handler(
                object_getter=object_getter,
                lookup_param=lookup_param,
                custom_handler=custom_handler,
                status_code=status_code,
            ),
        )
        return route.delete(
            working_path,
            url_name=url_name,
            response={status_code: str},
            description=description,
            operation_id=operation_id,
            summary=summary,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
        )(delete_item)

    @classmethod
    def _list_handler(
        cls, *, queryset_getter: t.Optional[t.Callable[..., QuerySet]]
    ) -> t.Callable:
        def list_items(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            """List Items of testing"""
            if queryset_getter:
                return queryset_getter(self, **kwargs)
            return self.service.get_all(**kwargs)

        list_items.__name__ = cls._change_name("list_items")
        return list_items

    @classmethod
    def _delete_handler(
        cls,
        *,
        object_getter: t.Optional[t.Callable[..., DjangoModel]],
        lookup_param: str,
        custom_handler: t.Optional[t.Callable[..., t.Any]],
        status_code: int,
    ) -> t.Callable:
        def delete_item(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one(pk=pk, **kwargs)
            )
            if not obj:  # pragma: no cover
                raise NotFound()
            self.check_object_permissions(obj)
            custom_handler(
                self, instance=obj, **kwargs
            ) if custom_handler else self.service.delete(instance=obj, **kwargs)
            return self.create_response(message="", status_code=status_code)

        delete_item.__name__ = cls._change_name("delete_item")
        return delete_item

    @classmethod
    def _find_one_handler(
        cls,
        *,
        object_getter: t.Optional[t.Callable[..., DjangoModel]],
        lookup_param: str,
    ) -> t.Callable:
        def get_item(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one(pk=pk, **kwargs)
            )
            if not obj:  # pragma: no cover
                raise NotFound()
            self.check_object_permissions(obj)
            return obj

        get_item.__name__ = cls._change_name("get_item")
        return get_item

    @classmethod
    def _patch_handler(
        cls,
        *,
        schema_in: t.Type[PydanticModel],
        object_getter: t.Optional[t.Callable[..., DjangoModel]],
        lookup_param: str,
        custom_handler: t.Optional[t.Callable[..., t.Any]],
    ) -> t.Callable:
        def patch_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one(pk=pk, **kwargs)
            )
            if not obj:  # pragma: no cover
                raise NotFound()
            self.check_object_permissions(obj)
            instance = (
                custom_handler(self, instance=obj, schema=data, **kwargs)
                if custom_handler
                else self.service.patch(instance=obj, schema=data, **kwargs)
            )
            assert instance, "`service.patch()` or `custom_handler` must return a value"
            return instance

        patch_item.__name__ = cls._change_name("patch_item")
        return patch_item

    @classmethod
    def _update_handler(
        cls,
        *,
        schema_in: t.Type[PydanticModel],
        object_getter: t.Optional[t.Callable[..., DjangoModel]],
        lookup_param: str,
        custom_handler: t.Optional[t.Callable[..., t.Any]],
    ) -> t.Callable:
        def update_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one(pk=pk, **kwargs)
            )

            if not obj:  # pragma: no cover
                raise NotFound()

            self.check_object_permissions(obj)
            instance = (
                custom_handler(self, instance=obj, schema=data, **kwargs)
                if custom_handler
                else self.service.update(instance=obj, schema=data, **kwargs)
            )
            assert instance, "`service.update` or `custom_handler` must return a value"
            return instance

        update_item.__name__ = cls._change_name("update_item")
        return update_item

    @classmethod
    def _create_handler(
        cls,
        *,
        schema_in: t.Type[PydanticModel],
        custom_handler: t.Optional[t.Callable[..., t.Any]],
    ) -> t.Callable:
        def create_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            instance = (
                custom_handler(self, data, **kwargs)
                if custom_handler
                else self.service.create(data, **kwargs)
            )
            assert instance, "`service.create` or  `custom_handler` must return a value"
            return instance

        create_item.__name__ = cls._change_name("create_item")
        return create_item


class ModelAsyncEndpointFactory(ModelEndpointFactory):
    """
    Factory for creating asynchronous CRUD operations of a model controller
    and for adding custom asynchronous route functions to controllers

    example:
    ```python

    api_controller
    class SampleMocelController(ModelControllerBase):

        create_sample = ModelAsyncEndpointFactory.create()
        update_sample = ModelAsyncEndpointFactory.update()

        delete_sample = ModelAsyncEndpointFactory.delete()
        patch_sample = ModelAsyncEndpointFactory.patch()

        get_sample = ModelAsyncEndpointFactory.get()
        list_samples = ModelAsyncEndpointFactory.list()
    ```
    """

    @classmethod
    def _list_handler(
        cls, *, queryset_getter: t.Optional[t.Callable[..., QuerySet]]
    ) -> t.Callable:
        async def list_items(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            """List Items of testing"""
            if queryset_getter:
                res = queryset_getter(self, **kwargs)
            else:
                res = self.service.get_all_async(**kwargs)  # type:ignore[assignment]

            return await check_if_coroutine(res)

        list_items.__name__ = cls._change_name("list_items")
        return list_items

    @classmethod
    def _delete_handler(
        cls,
        *,
        object_getter: t.Optional[t.Callable[..., DjangoModel]],
        lookup_param: str,
        custom_handler: t.Optional[t.Callable[..., t.Any]],
        status_code: int,
    ) -> t.Callable:
        async def delete_item(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one_async(pk=pk, **kwargs)
            )
            obj = await check_if_coroutine(obj)
            if not obj:  # pragma: no cover
                raise NotFound()
            self.check_object_permissions(obj)

            res = (
                custom_handler(self, instance=obj, **kwargs)
                if custom_handler
                else self.service.delete_async(instance=obj, **kwargs)  # type:ignore[arg-type]
            )

            await check_if_coroutine(res)
            return self.create_response(message="", status_code=status_code)

        delete_item.__name__ = cls._change_name("delete_item")
        return delete_item

    @classmethod
    def _find_one_handler(
        cls,
        *,
        object_getter: t.Optional[t.Callable[..., DjangoModel]],
        lookup_param: str,
    ) -> t.Callable:
        async def get_item(self: "ModelControllerBase", **kwargs: t.Any) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one_async(pk=pk, **kwargs)
            )
            obj = await check_if_coroutine(obj)

            if not obj:  # pragma: no cover
                raise NotFound()

            self.check_object_permissions(obj)
            return obj

        get_item.__name__ = cls._change_name("get_item")
        return get_item

    @classmethod
    def _patch_handler(
        cls,
        *,
        schema_in: t.Type[PydanticModel],
        object_getter: t.Optional[t.Callable[..., DjangoModel]],
        lookup_param: str,
        custom_handler: t.Optional[t.Callable[..., t.Any]],
    ) -> t.Callable:
        async def patch_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one_async(pk=pk, **kwargs)
            )
            obj = await check_if_coroutine(obj)

            if not obj:  # pragma: no cover
                raise NotFound()
            self.check_object_permissions(obj)

            instance = (
                custom_handler(self, instance=obj, schema=data, **kwargs)
                if custom_handler
                else self.service.patch_async(instance=obj, schema=data, **kwargs)  # type:ignore[arg-type]
            )
            instance = await check_if_coroutine(instance)

            assert (
                instance
            ), "`service.patch_async()` or `custom_handler` must return a value"
            return instance

        patch_item.__name__ = cls._change_name("patch_item")
        return patch_item

    @classmethod
    def _update_handler(
        cls,
        *,
        schema_in: t.Type[PydanticModel],
        object_getter: t.Optional[t.Callable[..., DjangoModel]],
        lookup_param: str,
        custom_handler: t.Optional[t.Callable[..., t.Any]],
    ) -> t.Callable:
        async def update_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            pk = kwargs.pop(lookup_param)
            obj = (
                object_getter(self, pk=pk, **kwargs)
                if object_getter
                else self.service.get_one_async(pk=pk, **kwargs)
            )
            obj = await check_if_coroutine(obj)

            if not obj:  # pragma: no cover
                raise NotFound()

            self.check_object_permissions(obj)
            instance = (
                custom_handler(self, instance=obj, schema=data, **kwargs)
                if custom_handler
                else self.service.update_async(instance=obj, schema=data, **kwargs)  # type:ignore[arg-type]
            )
            instance = await check_if_coroutine(instance)

            assert (
                instance
            ), "`service.update_async` or `custom_handler` must return a value"
            return instance

        update_item.__name__ = cls._change_name("update_item")
        return update_item

    @classmethod
    def _create_handler(
        cls,
        *,
        schema_in: t.Type[PydanticModel],
        custom_handler: t.Optional[t.Callable[..., t.Any]],
    ) -> t.Callable:
        async def create_item(
            self: "ModelControllerBase",
            data: schema_in = Body(default=...),  # type:ignore[valid-type]
            **kwargs: t.Any,
        ) -> t.Any:
            instance = (
                custom_handler(self, data, **kwargs)
                if custom_handler
                else self.service.create_async(data, **kwargs)
            )
            instance = await check_if_coroutine(instance)

            assert (
                instance
            ), "`service.create_async` or  `custom_handler` must return a value"
            return instance

        create_item.__name__ = cls._change_name("create_item")
        return create_item



"""Django Ninja Extra - Class Based Utility and more for Django Ninja(Fast Django REST framework)"""

__version__ = "0.21.1"

import django

from ninja_extra.controllers import (
    ControllerBase,
    ModelAsyncEndpointFactory,
    ModelConfig,
    ModelControllerBase,
    ModelControllerBuilder,
    ModelEndpointFactory,
    ModelPagination,
    ModelSchemaConfig,
    ModelService,
    ModelServiceBase,
    api_controller,
    http_delete,
    http_generic,
    http_get,
    http_patch,
    http_post,
    http_put,
)
from ninja_extra.controllers.route import route
from ninja_extra.dependency_resolver import get_injector, service_resolver
from ninja_extra.main import NinjaExtraAPI
from ninja_extra.pagination import paginate
from ninja_extra.router import Router
from ninja_extra.throttling import throttle

if django.VERSION < (3, 2):  # pragma: no cover
    default_app_config = "ninja_extra.apps.NinjaExtraConfig"


__all__ = [
    "ControllerBase",
    "api_controller",
    "NinjaExtraAPI",
    "route",
    "http_patch",
    "http_get",
    "http_put",
    "http_post",
    "http_delete",
    "http_generic",
    "permissions",
    "exceptions",
    "status",
    "shortcuts",
    "get_injector",
    "service_resolver",
    "lazy",
    "Router",
    "throttle",
    "paginate",
    "ModelControllerBase",
    "ModelConfig",
    "ModelService",
    "ModelSchemaConfig",
    "ModelControllerBuilder",
    "ModelPagination",
    "ModelServiceBase",
    "ModelControllerBase",
    "ModelEndpointFactory",
    "ModelAsyncEndpointFactory",
]



"""
DRF Exceptions
"""

import math
from typing import Any, Dict, List, Optional, Type, Union, no_type_check

from django.http import HttpRequest, HttpResponse, JsonResponse
from django.utils.encoding import force_str
from django.utils.translation import gettext_lazy as _
from django.utils.translation import ngettext
from ninja.errors import HttpError

from ninja_extra import status


@no_type_check
def _get_error_details(
    data: Union[List, Dict, "ErrorDetail"],
    default_code: Optional[Union[str, int]] = None,
) -> Union[List["ErrorDetail"], "ErrorDetail", Dict[Any, "ErrorDetail"]]:
    """
    Descend into a nested data structure, forcing any
    lazy translation strings or strings into `ErrorDetail`.
    """
    if isinstance(data, list):
        ret = [_get_error_details(item, default_code) for item in data]
        return ret
    elif isinstance(data, dict):
        ret = {
            key: _get_error_details(value, default_code) for key, value in data.items()
        }
        return ret

    text = force_str(data)
    code = getattr(data, "code", default_code)
    return ErrorDetail(text, code)


@no_type_check
def _get_codes(detail: Union[List, Dict, "ErrorDetail"]) -> Union[str, Dict]:
    if isinstance(detail, list):
        return [_get_codes(item) for item in detail]
    elif isinstance(detail, dict):
        return {key: _get_codes(value) for key, value in detail.items()}
    return detail.code


@no_type_check
def _get_full_details(detail: Union[List, Dict, "ErrorDetail"]) -> Dict:
    if isinstance(detail, list):
        return [_get_full_details(item) for item in detail]
    elif isinstance(detail, dict):
        return {key: _get_full_details(value) for key, value in detail.items()}
    return {"message": detail, "code": detail.code}


class ErrorDetail(str):
    """
    A string-like object that can additionally have a code.
    """

    code = None

    def __new__(
        cls, string: str, code: Optional[Union[str, int]] = None
    ) -> "ErrorDetail":
        self = super().__new__(cls, string)
        self.code = code
        return self

    def __eq__(self, other: object) -> bool:
        r = super().__eq__(other)
        try:
            return r and self.code == other.code  # type: ignore
        except AttributeError:
            return r

    def __ne__(self, other: object) -> bool:
        return not self.__eq__(other)

    def __repr__(self) -> str:
        return "ErrorDetail(string=%r, code=%r)" % (
            str(self),
            self.code,
        )

    def __hash__(self) -> Any:
        return hash(str(self))


class APIException(HttpError):
    """
    Base class for Django-Ninja-Extra exceptions.
    Subclasses should provide `.status_code` and `.default_detail` properties.
    """

    status_code = status.HTTP_500_INTERNAL_SERVER_ERROR
    default_detail = _("A server error occurred.")
    default_code = "error"

    def __init__(
        self,
        detail: Optional[Union[List, Dict, "ErrorDetail", str]] = None,
        code: Optional[Union[str, int]] = None,
    ) -> None:
        if detail is None:
            detail = force_str(self.default_detail)
        if code is None:
            code = self.default_code

        self.detail = _get_error_details(detail, code)

    def __str__(self) -> str:
        return str(self.detail)

    def get_codes(self) -> Union[str, Dict[Any, Any]]:
        """
        Return only the code part of the error details.

        Eg. {"name": ["required"]}
        """
        return _get_codes(self.detail)  # type: ignore

    def get_full_details(self) -> Dict[Any, Any]:
        """
        Return both the message & code parts of the error details.

        Eg. {"name": [{"message": "This field is required.", "code": "required"}]}
        """
        return _get_full_details(self.detail)  # type: ignore


# The recommended style for using `ValidationError`,
# in order to minimize potential confusion with Django's
# built in `ValidationError`. For example:
#
# from ninja_extra import exceptions
# raise exceptions.ValidationError('Value was invalid')


class ValidationError(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _("Invalid input.")
    default_code = "invalid"

    def __init__(
        self,
        detail: Optional[Union[List, Dict, "ErrorDetail", str]] = None,
        code: Optional[Union[str, int]] = None,
    ):
        if detail is None:
            detail = force_str(self.default_detail)
        if code is None:
            code = self.default_code

        # For validation failures, we may collect many errors together,
        # so the details should always be coerced to a list if not already.
        if not isinstance(detail, dict) and not isinstance(detail, list):
            detail = [detail]

        self.detail = _get_error_details(detail, code)


class ParseError(APIException):
    status_code = status.HTTP_400_BAD_REQUEST
    default_detail = _("Malformed request.")
    default_code = "parse_error"


class AuthenticationFailed(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = _("Incorrect authentication credentials.")
    default_code = "authentication_failed"


class NotAuthenticated(APIException):
    status_code = status.HTTP_401_UNAUTHORIZED
    default_detail = _("Authentication credentials were not provided.")
    default_code = "not_authenticated"


class PermissionDenied(APIException):
    status_code = status.HTTP_403_FORBIDDEN
    default_detail = _("You do not have permission to perform this action.")
    default_code = "permission_denied"


class NotFound(APIException):
    status_code = status.HTTP_404_NOT_FOUND
    default_detail = _("Not found.")
    default_code = "not_found"


class MethodNotAllowed(APIException):
    status_code = status.HTTP_405_METHOD_NOT_ALLOWED
    default_detail = _('Method "{method}" not allowed.')
    default_code = "method_not_allowed"

    def __init__(
        self,
        method: str,
        detail: Optional[Union[List, Dict, "ErrorDetail", str]] = None,
        code: Optional[Union[str, int]] = None,
    ):
        if detail is None:
            detail = force_str(self.default_detail).format(method=method)
        super().__init__(detail, code)


class NotAcceptable(APIException):
    status_code = status.HTTP_406_NOT_ACCEPTABLE
    default_detail = _("Could not satisfy the request Accept header.")
    default_code = "not_acceptable"

    def __init__(
        self,
        detail: Optional[Union[List, Dict, "ErrorDetail"]] = None,
        code: Optional[Union[str, int]] = None,
        available_renderers: Optional[str] = None,
    ):
        self.available_renderers = available_renderers
        super().__init__(detail, code)


class UnsupportedMediaType(APIException):
    status_code = status.HTTP_415_UNSUPPORTED_MEDIA_TYPE
    default_detail = _('Unsupported media type "{media_type}" in request.')
    default_code = "unsupported_media_type"

    def __init__(
        self,
        media_type: str,
        detail: Optional[Union[List, Dict, "ErrorDetail", str]] = None,
        code: Optional[Union[str, int]] = None,
    ):
        if detail is None:
            detail = force_str(self.default_detail).format(media_type=media_type)
        super().__init__(detail, code)


class Throttled(APIException):
    status_code = status.HTTP_429_TOO_MANY_REQUESTS
    default_detail = _("Request was throttled.")
    extra_detail_singular = _("Expected available in {wait} second.")
    extra_detail_plural = _("Expected available in {wait} seconds.")
    default_code = "throttled"

    def __init__(
        self,
        wait: Optional[float] = None,
        detail: Optional[Any] = None,
        code: Optional[Any] = None,
    ) -> None:
        if detail is None:
            detail = force_str(self.default_detail)
        if wait is not None:
            wait = math.ceil(wait)
            detail = " ".join(
                (
                    detail,
                    force_str(
                        ngettext(
                            self.extra_detail_singular.format(wait=wait),
                            self.extra_detail_plural.format(wait=wait),
                            wait,
                        )
                    ),
                )
            )
        self.wait = wait
        super().__init__(detail, code)


def server_error(request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
    """
    Generic 500 error handler.
    """
    data = {"error": "Server Error (500)"}
    return JsonResponse(data, status=status.HTTP_500_INTERNAL_SERVER_ERROR)


def bad_request(
    request: HttpRequest, exception: Type[APIException], *args: Any, **kwargs: Any
) -> HttpResponse:
    """
    Generic 400 error handler.
    """
    data = {"error": "Bad Request (400)"}
    return JsonResponse(data, status=status.HTTP_400_BAD_REQUEST)



from importlib import import_module
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from django.core.exceptions import ImproperlyConfigured
from django.http import HttpRequest, HttpResponse
from django.urls import URLPattern, URLResolver
from django.utils.module_loading import import_string, module_has_submodule
from ninja import NinjaAPI
from ninja.constants import NOT_SET
from ninja.openapi.docs import DocsBase, Swagger
from ninja.parser import Parser
from ninja.renderers import BaseRenderer
from ninja.throttling import BaseThrottle
from ninja.types import DictStrAny, TCallable

from ninja_extra import exceptions, router
from ninja_extra.compatible import NOT_SET_TYPE
from ninja_extra.controllers.base import APIController, ControllerBase
from ninja_extra.controllers.registry import ControllerRegistry

__all__ = [
    "NinjaExtraAPI",
]


class NinjaExtraAPI(NinjaAPI):
    def __init__(
        self,
        *,
        title: str = "NinjaExtraAPI",
        version: str = "1.0.0",
        description: str = "",
        openapi_url: Optional[str] = "/openapi.json",
        docs: DocsBase = Swagger(),
        docs_url: Optional[str] = "/docs",
        docs_decorator: Optional[Callable[[TCallable], TCallable]] = None,
        servers: Optional[List[DictStrAny]] = None,
        urls_namespace: Optional[str] = None,
        csrf: bool = False,
        auth: Optional[Union[Sequence[Callable], Callable, NOT_SET_TYPE]] = NOT_SET,
        throttle: Union[BaseThrottle, List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        renderer: Optional[BaseRenderer] = None,
        parser: Optional[Parser] = None,
        openapi_extra: Optional[Dict[str, Any]] = None,
        app_name: str = "ninja",
        **kwargs: Any,
    ) -> None:
        super(NinjaExtraAPI, self).__init__(
            title=title,
            version=version,
            description=description,
            openapi_url=openapi_url,
            docs_url=docs_url,
            urls_namespace=urls_namespace,
            csrf=csrf,
            auth=auth,
            renderer=renderer,
            parser=parser,
            openapi_extra=openapi_extra,
            servers=servers,
            docs=docs,
            docs_decorator=docs_decorator,
            **kwargs,
        )
        self.app_name = app_name
        self.exception_handler(exceptions.APIException)(self.api_exception_handler)
        self._routers: List[Tuple[str, router.Router]] = []  # type: ignore
        self.default_router = router.Router()
        self.add_router("", self.default_router)

    def api_exception_handler(
        self, request: HttpRequest, exc: exceptions.APIException
    ) -> HttpResponse:
        headers: Dict = {}
        if isinstance(exc, exceptions.Throttled):
            headers["Retry-After"] = "%d" % float(exc.wait or 0.0)

        if isinstance(exc.detail, (list, dict)):
            data = exc.detail
        else:
            data = {"detail": exc.detail}

        response = self.create_response(request, data, status=exc.status_code)
        for k, v in headers.items():
            response.setdefault(k, v)

        return response

    @property
    def urls(self) -> Tuple[List[Union[URLResolver, URLPattern]], str, str]:
        _url_tuple = super().urls
        return (
            _url_tuple[0],
            self.app_name,
            str(_url_tuple[len(_url_tuple) - 1]),
        )

    def register_controllers(
        self, *controllers: Union[Type[ControllerBase], Type, str]
    ) -> None:
        for controller in controllers:
            if isinstance(controller, str):
                controller = cast(
                    Union[Type[ControllerBase], Type], import_string(controller)
                )

            if not issubclass(controller, ControllerBase):
                raise ImproperlyConfigured(
                    f"{controller.__class__.__name__} class is not a controller"
                )
            api_controller: APIController = controller.get_api_controller()
            if not api_controller.registered:
                self._routers.extend(api_controller.build_routers())  # type: ignore
                api_controller.set_api_instance(self)
                api_controller.registered = True

    def auto_discover_controllers(self) -> None:
        from django.apps import apps

        installed_apps = [
            v for k, v in apps.app_configs.items() if not v.name.startswith("django.")
        ]
        possible_module_name = ["api", "controllers"]

        for app_module in installed_apps:
            try:
                app_module_ = import_module(app_module.name)
                for module in possible_module_name:
                    if module_has_submodule(app_module_, module):
                        mod_path = "%s.%s" % (app_module.name, module)
                        import_module(mod_path)
                self.register_controllers(
                    *ControllerRegistry.get_controllers().values()
                )
            except ImportError as ex:  # pragma: no cover
                raise ex



from abc import ABC, abstractmethod
from typing import Any, Optional

from django.http import HttpRequest
from ninja.security.apikey import APIKeyBase, APIKeyCookie, APIKeyHeader, APIKeyQuery

__all__ = [
    "AsyncAPIKeyBase",
    "AsyncAPIKeyQuery",
    "AsyncAPIKeyCookie",
    "AsyncAPIKeyHeader",
]


class AsyncAPIKeyBase(APIKeyBase, ABC):
    async def __call__(self, request: HttpRequest) -> Optional[Any]:
        key = self._get_key(request)
        return await self.authenticate(request, key)

    @abstractmethod
    async def authenticate(
        self, request: HttpRequest, key: Optional[str]
    ) -> Optional[Any]:
        pass  # pragma: no cover


class AsyncAPIKeyQuery(AsyncAPIKeyBase, APIKeyQuery, ABC):
    pass


class AsyncAPIKeyCookie(AsyncAPIKeyBase, APIKeyCookie, ABC):
    pass


class AsyncAPIKeyHeader(AsyncAPIKeyBase, APIKeyHeader, ABC):
    pass



from typing import Any, List, Type, Union

from ninja_extra.permissions.base import (
    BasePermission,
    OperandHolder,
    SingleOperandHolder,
)

PermissionType = List[
    Union[Type[BasePermission], OperandHolder, SingleOperandHolder, BasePermission, Any]
]



from functools import wraps
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Union,
    cast,
)

from asgiref.sync import sync_to_async
from django.http import HttpRequest
from ninja.pagination import PaginationBase

from ninja_extra.controllers.route.context import RouteContext
from ninja_extra.shortcuts import add_ninja_contribute_args

if TYPE_CHECKING:  # pragma: no cover
    from ninja_extra.controllers import ControllerBase


class PaginatorOperation:
    def __init__(
        self,
        *,
        paginator: PaginationBase,
        view_func: Callable,
        paginator_kwargs_name: str = "pagination",
    ) -> None:
        self.paginator = paginator
        self.paginator_kwargs_name = paginator_kwargs_name
        self.view_func = view_func

        paginator_view = self.get_view_function()
        self.as_view = wraps(view_func)(paginator_view)
        add_ninja_contribute_args(
            self.as_view,
            (
                self.paginator_kwargs_name,
                self.paginator.Input,
                self.paginator.InputSource,
            ),
        )
        paginator_view.paginator_operation = self  # type:ignore[attr-defined]

    @property
    def view_func_has_kwargs(self) -> bool:  # pragma: no cover
        return self.paginator.pass_parameter is not None

    def get_view_function(self) -> Callable:
        def as_view(
            request_or_controller: Union["ControllerBase", HttpRequest],
            *args: Any,
            **kw: Any,
        ) -> Any:
            func_kwargs = dict(**kw)
            pagination_params = func_kwargs.pop(self.paginator_kwargs_name)
            if self.paginator.pass_parameter:
                func_kwargs[self.paginator.pass_parameter] = pagination_params

            items = self.view_func(request_or_controller, *args, **func_kwargs)
            if hasattr(request_or_controller, "context") and isinstance(
                request_or_controller.context, RouteContext
            ):
                request = request_or_controller.context.request
                assert request, "Request object is None"
            else:
                request = request_or_controller
            params = dict(kw)
            params["request"] = request
            return self.paginator.paginate_queryset(items, **params)

        return as_view


class AsyncPaginatorOperation(PaginatorOperation):
    def get_view_function(self) -> Callable:
        async def as_view(
            request_or_controller: Union["ControllerBase", HttpRequest],
            *args: Any,
            **kw: Any,
        ) -> Any:
            func_kwargs = dict(**kw)
            pagination_params = func_kwargs.pop(self.paginator_kwargs_name)
            if self.paginator.pass_parameter:
                func_kwargs[self.paginator.pass_parameter] = pagination_params

            items = await self.view_func(request_or_controller, *args, **func_kwargs)
            if hasattr(request_or_controller, "context") and isinstance(
                request_or_controller.context, RouteContext
            ):
                request = request_or_controller.context.request
                assert request, "Request object is None"
            else:
                request = request_or_controller
            params = dict(kw)
            params["request"] = request
            paginate_queryset = cast(
                Callable, sync_to_async(self.paginator.paginate_queryset)
            )
            return await paginate_queryset(items, **params)

        return as_view



import typing as t

from ninja.constants import NOT_SET, NOT_SET_TYPE
from ninja.signature import is_async
from ninja.throttling import BaseThrottle
from ninja.types import TCallable

from ninja_extra.constants import (
    DELETE,
    GET,
    PATCH,
    POST,
    PUT,
    ROUTE_FUNCTION,
    ROUTE_METHODS,
)
from ninja_extra.permissions import BasePermission
from ninja_extra.schemas import RouteParameter

from .route_functions import AsyncRouteFunction, RouteFunction


class RouteInvalidParameterException(Exception):
    pass


class Route(object):
    """
    Decorates Controller class methods with HTTP Operation methods and as a route function handler

    Example:
        ```python
        from ninja_extra import api_controller

        @api_controller
        class SampleController:
            @route.post('/create')
            def create_sample():
                pass

            @route.get('/list')
            def get_all_samples():
                pass
        ```
    """

    permissions: t.Optional[t.Optional[t.List[t.Type[BasePermission]]]] = None

    def __init__(
        self,
        view_func: TCallable,
        *,
        path: str,
        methods: t.List[str],
        auth: t.Any = NOT_SET,
        throttle: t.Union[BaseThrottle, t.List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: t.Union[t.Any, t.List[t.Any]] = NOT_SET,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: t.Optional[str] = None,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> None:
        if not isinstance(methods, list):
            raise RouteInvalidParameterException("methods must be a list")

        methods = [m.upper() for m in methods]
        not_valid_methods = list(set(methods) - set(ROUTE_METHODS))
        if not_valid_methods:
            raise RouteInvalidParameterException(
                f"Method {','.join(not_valid_methods)} not allowed"
            )

        _response = response
        if isinstance(response, list):
            _response_computed = {}
            for item in response:
                if isinstance(item, dict):
                    _response_computed.update(item)
                elif isinstance(item, tuple):
                    _response_computed.update({item[0]: item[1]})
            if not _response_computed:
                raise RouteInvalidParameterException(
                    f"Invalid response configuration: {response}"
                )
            _response = _response_computed

        ninja_route_params = RouteParameter(
            path=path,
            methods=methods,
            auth=auth,
            throttle=throttle,
            response=_response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            openapi_extra=openapi_extra,
        )
        self.route_params = ninja_route_params
        self.is_async = is_async(view_func)
        self.permissions = permissions  # type: ignore[assignment]
        self.view_func = view_func

    @classmethod
    def _create_route_function(
        cls,
        view_func: TCallable,
        *,
        path: str,
        methods: t.List[str],
        auth: t.Any = NOT_SET,
        throttle: t.Union[BaseThrottle, t.List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: t.Union[t.Any, t.List[t.Any]] = NOT_SET,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: t.Optional[str] = None,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> TCallable:
        if response is NOT_SET:
            type_hint = t.get_type_hints(view_func).get("return") or NOT_SET
            if not isinstance(type_hint, t._SpecialForm):
                response = type_hint
        route_obj = cls(
            view_func,
            path=path,
            methods=methods,
            auth=auth,
            response=response,
            operation_id=operation_id,
            summary=summary,
            description=description,
            tags=tags,
            deprecated=deprecated,
            by_alias=by_alias,
            exclude_unset=exclude_unset,
            exclude_defaults=exclude_defaults,
            exclude_none=exclude_none,
            url_name=url_name,
            include_in_schema=include_in_schema,
            permissions=permissions,
            openapi_extra=openapi_extra,
            throttle=throttle,
        )
        route_function_class = RouteFunction
        if route_obj.is_async:
            route_function_class = AsyncRouteFunction

        setattr(view_func, ROUTE_FUNCTION, route_function_class(route=route_obj))
        return view_func

    @classmethod
    def get(
        cls,
        path: str = "",
        *,
        auth: t.Any = NOT_SET,
        throttle: t.Union[BaseThrottle, t.List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: t.Union[t.Any, t.List[t.Any]] = NOT_SET,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: t.Optional[str] = None,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable[[TCallable], TCallable]:
        """
        A GET Operation method decorator
         eg.

        ```python
        @http_get()
        def get_operation(self):
        ...
        ```
        :param path: uniques endpoint path string
        :param auth: endpoint authentication method. default: `NOT_SET`
        :param response: `dict[status_code, schema]` or `Schema` used validated returned response. default: `None`
        :param operation_id: unique id that distinguishes `operation` in path view. default: `None`
        :param summary: describes your endpoint. default: `None`
        :param description: other description of your endpoint. default: `None`
        :param tags: list of strings for grouping endpoints only for documentation purpose. default: `None`
        :param deprecated: declares an endpoint deprecated. default: `None`
        :param by_alias: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_unset: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_defaults: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_none: pydantic schema filters applied to `response` schema object. default: `False`
        :param url_name: a name to an endpoint which can be resolved using `reverse` function in django. default: `None`
        :param include_in_schema: indicates whether an endpoint should appear on the swagger documentation
        :param permissions: collection permission classes. default: `None`
        :return: Route[GET]
        """

        def decorator(view_func: TCallable) -> TCallable:
            return cls._create_route_function(
                view_func,
                path=path,
                methods=[GET],
                auth=auth,
                response=response,
                operation_id=operation_id,
                summary=summary,
                description=description,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                url_name=url_name,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
                throttle=throttle,
            )

        return decorator

    @classmethod
    def post(
        cls,
        path: str = "",
        *,
        auth: t.Any = NOT_SET,
        throttle: t.Union[BaseThrottle, t.List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: t.Union[t.Any, t.List[t.Any]] = NOT_SET,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: t.Optional[str] = None,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable[[TCallable], TCallable]:
        """
        A POST Operation method decorator
        eg.

        ```python
         @http_post()
         def post_operation(self,  create_schema: Schema):
            ...
        ```
        :param path: uniques endpoint path string
        :param auth: endpoint authentication method. default: `NOT_SET`
        :param response: `dict[status_code, schema]` or `Schema` used validated returned response. default: `None`
        :param operation_id: unique id that distinguishes `operation` in path view. default: `None`
        :param summary: describes your endpoint. default: `None`
        :param description: other description of your endpoint. default: `None`
        :param tags: list of strings for grouping endpoints only for documentation purpose. default: `None`
        :param deprecated: declares an endpoint deprecated. default: `None`
        :param by_alias: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_unset: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_defaults: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_none: pydantic schema filters applied to `response` schema object. default: `False`
        :param url_name: a name to an endpoint which can be resolved using `reverse` function in django. default: `None`
        :param include_in_schema: indicates whether an endpoint should appear on the swagger documentation
        :param permissions: collection permission classes. default: `None`
        :return: Route[POST]
        """

        def decorator(view_func: TCallable) -> TCallable:
            return cls._create_route_function(
                view_func,
                path=path,
                methods=[POST],
                auth=auth,
                response=response,
                operation_id=operation_id,
                summary=summary,
                description=description,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                url_name=url_name,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
                throttle=throttle,
            )

        return decorator

    @classmethod
    def delete(
        cls,
        path: str = "",
        *,
        auth: t.Any = NOT_SET,
        throttle: t.Union[BaseThrottle, t.List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: t.Union[t.Any, t.List[t.Any]] = NOT_SET,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: t.Optional[str] = None,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable[[TCallable], TCallable]:
        """
        A DELETE Operation method decorator
        eg.

        ```python
        @http_delete('/{int:some_id}')
        def delete_operation(self, some_id: int):
            ...
        ```
        :param path: uniques endpoint path string
        :param auth: endpoint authentication method. default: `NOT_SET`
        :param response: `dict[status_code, schema]` or `Schema` used validated returned response. default: `None`
        :param operation_id: unique id that distinguishes `operation` in path view. default: `None`
        :param summary: describes your endpoint. default: `None`
        :param description: other description of your endpoint. default: `None`
        :param tags: list of strings for grouping endpoints only for documentation purpose. default: `None`
        :param deprecated: declares an endpoint deprecated. default: `None`
        :param by_alias: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_unset: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_defaults: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_none: pydantic schema filters applied to `response` schema object. default: `False`
        :param url_name: a name to an endpoint which can be resolved using `reverse` function in django. default: `None`
        :param include_in_schema: indicates whether an endpoint should appear on the swagger documentation
        :param permissions: collection permission classes. default: `None`
        :return: Route[DELETE]
        """

        def decorator(view_func: TCallable) -> TCallable:
            return cls._create_route_function(
                view_func,
                path=path,
                methods=[DELETE],
                auth=auth,
                response=response,
                operation_id=operation_id,
                summary=summary,
                description=description,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                url_name=url_name,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
                throttle=throttle,
            )

        return decorator

    @classmethod
    def patch(
        cls,
        path: str = "",
        *,
        auth: t.Any = NOT_SET,
        throttle: t.Union[BaseThrottle, t.List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: t.Union[t.Any, t.List[t.Any]] = NOT_SET,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: t.Optional[str] = None,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable[[TCallable], TCallable]:
        """
        A PATCH Operation method decorator
        eg.

        ```python

        @http_patch('/{int:some_id}')
        def patch_operation(self,  some_id: int):
            ...
        ```
        :param path: uniques endpoint path string
        :param auth: endpoint authentication method. default: `NOT_SET`
        :param response: `dict[status_code, schema]` or `Schema` used validated returned response. default: `None`
        :param operation_id: unique id that distinguishes `operation` in path view. default: `None`
        :param summary: describes your endpoint. default: `None`
        :param description: other description of your endpoint. default: `None`
        :param tags: list of strings for grouping endpoints only for documentation purpose. default: `None`
        :param deprecated: declares an endpoint deprecated. default: `None`
        :param by_alias: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_unset: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_defaults: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_none: pydantic schema filters applied to `response` schema object. default: `False`
        :param url_name: a name to an endpoint which can be resolved using `reverse` function in django. default: `None`
        :param include_in_schema: indicates whether an endpoint should appear on the swagger documentation
        :param permissions: collection permission classes. default: `None`
        :return: Route[PATCH]
        """

        def decorator(view_func: TCallable) -> TCallable:
            return cls._create_route_function(
                view_func,
                path=path,
                methods=[PATCH],
                auth=auth,
                response=response,
                operation_id=operation_id,
                summary=summary,
                description=description,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                url_name=url_name,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
                throttle=throttle,
            )

        return decorator

    @classmethod
    def put(
        cls,
        path: str = "",
        *,
        auth: t.Any = NOT_SET,
        throttle: t.Union[BaseThrottle, t.List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: t.Union[t.Any, t.List[t.Any]] = NOT_SET,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: t.Optional[str] = None,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable[[TCallable], TCallable]:
        """
         A PUT Operation method decorator
        eg.

        ```python

        @http_put('/{int:some_id}')
        def put_operation(self, some_id: int):
            ...
        ```
         :param path: uniques endpoint path string
         :param auth: endpoint authentication method. default: `NOT_SET`
         :param response: `dict[status_code, schema]` or `Schema` used validated returned response. default: `None`
         :param operation_id: unique id that distinguishes `operation` in path view. default: `None`
         :param summary: describes your endpoint. default: `None`
         :param description: other description of your endpoint. default: `None`
         :param tags: list of strings for grouping endpoints only for documentation purpose. default: `None`
         :param deprecated: declares an endpoint deprecated. default: `None`
         :param by_alias: pydantic schema filters applied to `response` schema object. default: `False`
         :param exclude_unset: pydantic schema filters applied to `response` schema object. default: `False`
         :param exclude_defaults: pydantic schema filters applied to `response` schema object. default: `False`
         :param exclude_none: pydantic schema filters applied to `response` schema object. default: `False`
         :param url_name: a name to an endpoint which can be resolved using `reverse` function in django. default: `None`
         :param include_in_schema: indicates whether an endpoint should appear on the swagger documentation
         :param permissions: collection permission classes. default: `None`
         :return: Route[PUT]
        """

        def decorator(view_func: TCallable) -> TCallable:
            return cls._create_route_function(
                view_func,
                path=path,
                methods=[PUT],
                auth=auth,
                response=response,
                operation_id=operation_id,
                summary=summary,
                description=description,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                url_name=url_name,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
                throttle=throttle,
            )

        return decorator

    @classmethod
    def generic(
        cls,
        path: str = "",
        *,
        methods: t.List[str],
        auth: t.Any = NOT_SET,
        throttle: t.Union[BaseThrottle, t.List[BaseThrottle], NOT_SET_TYPE] = NOT_SET,
        response: t.Union[t.Any, t.List[t.Any]] = NOT_SET,
        operation_id: t.Optional[str] = None,
        summary: t.Optional[str] = None,
        description: t.Optional[str] = None,
        tags: t.Optional[t.List[str]] = None,
        deprecated: t.Optional[bool] = None,
        by_alias: bool = False,
        exclude_unset: bool = False,
        exclude_defaults: bool = False,
        exclude_none: bool = False,
        url_name: t.Optional[str] = None,
        include_in_schema: bool = True,
        permissions: t.Optional[
            t.List[t.Union[t.Type[BasePermission], BasePermission, t.Any]]
        ] = None,
        openapi_extra: t.Optional[t.Dict[str, t.Any]] = None,
    ) -> t.Callable[[TCallable], TCallable]:
        """
        A Custom Operation method decorator, for creating route with more than one operation
        eg.

        ```python

        @http_generic('', methods=['POST', 'GET'])
        def list_create(self, some_schema: Optional[Schema] = None):
           ...
        ```
        :param path: uniques endpoint path string
        :param methods: List of operations `GET, PUT, PATCH, DELETE, POST`
        :param auth: endpoint authentication method. default: `NOT_SET`
        :param response: `dict[status_code, schema]` or `Schema` used validated returned response. default: `None`
        :param operation_id: unique id that distinguishes `operation` in path view. default: `None`
        :param summary: describes your endpoint. default: `None`
        :param description: other description of your endpoint. default: `None`
        :param tags: list of strings for grouping endpoints only for documentation purpose. default: `None`
        :param deprecated: declares an endpoint deprecated. default: `None`
        :param by_alias: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_unset: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_defaults: pydantic schema filters applied to `response` schema object. default: `False`
        :param exclude_none: pydantic schema filters applied to `response` schema object. default: `False`
        :param url_name: a name to an endpoint which can be resolved using `reverse` function in django. default: `None`
        :param include_in_schema: indicates whether an endpoint should appear on the swagger documentation
        :param permissions: collection permission classes. default: `None`
        :return: Route[PATCH]
        """

        def decorator(view_func: TCallable) -> TCallable:
            return cls._create_route_function(
                view_func,
                path=path,
                methods=methods,
                auth=auth,
                response=response,
                operation_id=operation_id,
                summary=summary,
                description=description,
                tags=tags,
                deprecated=deprecated,
                by_alias=by_alias,
                exclude_unset=exclude_unset,
                exclude_defaults=exclude_defaults,
                exclude_none=exclude_none,
                url_name=url_name,
                include_in_schema=include_in_schema,
                permissions=permissions,
                openapi_extra=openapi_extra,
                throttle=throttle,
            )

        return decorator


route = Route

http_get = route.get
http_patch = route.patch

http_put = route.put
http_delete = route.delete

http_post = route.post
http_generic = route.generic



import inspect
import typing as t

from ninja_extra.constants import ROUTE_FUNCTION

if t.TYPE_CHECKING:  # pragma: no cover
    from ninja_extra.controllers import RouteFunction


def get_function_name(func_class: t.Any) -> str:
    if inspect.isfunction(func_class) or inspect.isclass(func_class):
        return str(func_class.__name__)
    return str(func_class.__class__.__name__)


@t.no_type_check
def get_route_function(func: t.Callable) -> t.Optional["RouteFunction"]:
    if hasattr(func, ROUTE_FUNCTION):
        return func.__dict__[ROUTE_FUNCTION]
    return None  # pragma: no cover



from .base import ControllerBase, ModelControllerBase, api_controller
from .model import (
    ModelAsyncEndpointFactory,
    ModelConfig,
    ModelControllerBuilder,
    ModelEndpointFactory,
    ModelPagination,
    ModelSchemaConfig,
    ModelService,
    ModelServiceBase,
)
from .response import Detail, Id, Ok
from .route import (
    Route,
    RouteInvalidParameterException,
    http_delete,
    http_generic,
    http_get,
    http_patch,
    http_post,
    http_put,
    route,
)
from .route.context import RouteContext
from .route.route_functions import AsyncRouteFunction, RouteFunction

__all__ = [
    "api_controller",
    "route",
    "http_get",
    "http_post",
    "http_put",
    "http_delete",
    "http_patch",
    "http_generic",
    "ControllerBase",
    "RouteInvalidParameterException",
    "AsyncRouteFunction",
    "RouteFunction",
    "Route",
    "Ok",
    "Id",
    "Detail",
    "RouteContext",
    "ModelControllerBase",
    "ModelConfig",
    "ModelService",
    "ModelSchemaConfig",
    "ModelControllerBuilder",
    "ModelPagination",
    "ModelServiceBase",
    "ModelEndpointFactory",
    "ModelAsyncEndpointFactory",
]



import logging
from abc import ABC, abstractmethod
from typing import Any, Optional

from django.conf import settings
from django.http import HttpRequest
from ninja.security.http import DecodeError, HttpBasicAuth, HttpBearer

logger = logging.getLogger("django")

__all__ = ["AsyncHttpBearer", "AsyncHttpBasicAuth"]


class AsyncHttpBearer(HttpBearer, ABC):
    async def __call__(self, request: HttpRequest) -> Optional[Any]:
        headers = request.headers
        auth_value = headers.get(self.header)
        if not auth_value:
            return None
        parts = auth_value.split(" ")

        if parts[0].lower() != self.openapi_scheme:
            if settings.DEBUG:
                logger.error(f"Unexpected auth - '{auth_value}'")
            return None
        token = " ".join(parts[1:])
        return await self.authenticate(request, token)

    @abstractmethod
    async def authenticate(self, request: HttpRequest, token: str) -> Optional[Any]:
        pass  # pragma: no cover


class AsyncHttpBasicAuth(HttpBasicAuth, ABC):
    async def __call__(self, request: HttpRequest) -> Optional[Any]:
        headers = request.headers
        auth_value = headers.get(self.header)
        if not auth_value:
            return None

        try:
            username, password = self.decode_authorization(auth_value)
        except DecodeError as e:
            if settings.DEBUG:
                logger.exception(e)
            return None
        return await self.authenticate(request, username, password)

    @abstractmethod
    async def authenticate(
        self, request: HttpRequest, username: str, password: str
    ) -> Optional[Any]:
        pass  # pragma: no cover



import sys

from ninja.constants import NOT_SET

__all__ = ["asynccontextmanager", "NOT_SET_TYPE"]

if sys.version_info >= (3, 7):  # pragma: no cover
    from contextlib import asynccontextmanager as asynccontextmanager
else:  # pragma: no cover
    from contextlib2 import asynccontextmanager as asynccontextmanager

try:
    from ninja.constants import NOT_SET_TYPE
except Exception:  # pragma: no cover
    NOT_SET_TYPE = type(NOT_SET)



import inspect
import logging
from abc import ABC, abstractmethod
from functools import wraps
from operator import attrgetter, itemgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    List,
    Optional,
    Tuple,
    Type,
    Union,
    cast,
    overload,
)

from asgiref.sync import sync_to_async
from django.db.models import QuerySet
from django.http import HttpRequest
from ninja import Field, P, Query, Schema
from ninja.constants import NOT_SET
from ninja.signature import is_async
from pydantic import BaseModel

from ninja_extra.conf import settings
from ninja_extra.shortcuts import add_ninja_contribute_args

logger = logging.getLogger()

if TYPE_CHECKING:  # pragma: no cover
    from .controllers import ControllerBase

__all__ = [
    "OrderingBase",
    "Ordering",
    "ordering",
    "OrderatorOperation",
    "AsyncOrderatorOperation",
]


class OrderingBase(ABC):
    class Input(Schema): ...

    InputSource = Query(...)

    def __init__(self, *, pass_parameter: Optional[str] = None, **kwargs: Any) -> None:
        self.pass_parameter = pass_parameter

    @abstractmethod
    def ordering_queryset(
        self, items: Union[QuerySet, List], ordering_input: Any
    ) -> Union[QuerySet, List]: ...


class Ordering(OrderingBase):
    class Input(Schema):
        ordering: Optional[str] = Field(None)

    def __init__(
        self,
        ordering_fields: Optional[List[str]] = None,
        pass_parameter: Optional[str] = None,
    ) -> None:
        super().__init__(pass_parameter=pass_parameter)
        self.ordering_fields = ordering_fields or "__all__"
        self.Input = self.create_input(ordering_fields)  # type:ignore

    def create_input(self, ordering_fields: Optional[List[str]]) -> Type[Input]:
        if ordering_fields:

            class DynamicInput(Ordering.Input):
                ordering: Query[Optional[str], P(default=",".join(ordering_fields))]  # type:ignore[type-arg,valid-type]

            return DynamicInput
        return Ordering.Input

    def ordering_queryset(
        self, items: Union[QuerySet, List], ordering_input: Input
    ) -> Union[QuerySet, List]:
        ordering_ = self.get_ordering(items, ordering_input.ordering)
        if ordering_:
            if isinstance(items, QuerySet):
                return items.order_by(*ordering_)
            elif isinstance(items, list) and items:

                def multisort(xs: List, specs: List[Tuple[str, bool]]) -> List:
                    orerator = itemgetter if isinstance(xs[0], dict) else attrgetter
                    for key, reverse in reversed(specs):
                        xs.sort(key=orerator(key), reverse=reverse)
                    return xs

                return multisort(
                    items,
                    [
                        (o[int(o.startswith("-")) :], o.startswith("-"))
                        for o in ordering_
                    ],
                )
        return items

    def get_ordering(
        self, items: Union[QuerySet, List], value: Optional[str]
    ) -> List[str]:
        if value:
            fields = [param.strip() for param in value.split(",")]
            return self.remove_invalid_fields(items, fields)
        return []

    def remove_invalid_fields(
        self, items: Union[QuerySet, List], fields: List[str]
    ) -> List[str]:
        valid_fields = list(self.get_valid_fields(items))

        def term_valid(term: str) -> bool:
            if term.startswith("-"):
                term = term[1:]
            return term in valid_fields

        return [term for term in fields if term_valid(term)]

    def get_valid_fields(self, items: Union[QuerySet, List]) -> List[str]:
        valid_fields: List[str] = []
        if self.ordering_fields == "__all__":
            if isinstance(items, QuerySet):
                valid_fields = self.get_all_valid_fields_from_queryset(items)
            elif isinstance(items, list):
                valid_fields = self.get_all_valid_fields_from_list(items)
        else:
            valid_fields = list(self.ordering_fields)
        return valid_fields

    def get_all_valid_fields_from_queryset(self, items: QuerySet) -> List[str]:
        return [str(field.name) for field in items.model._meta.fields] + [
            str(key) for key in items.query.annotations
        ]

    def get_all_valid_fields_from_list(self, items: List) -> List[str]:
        if not items:
            return []
        item = items[0]
        if isinstance(item, BaseModel):
            return list(item.model_fields.keys())
        if isinstance(item, dict):
            return list(item.keys())
        if hasattr(item, "_meta") and hasattr(item._meta, "fields"):
            return [str(field.name) for field in item._meta.fields]
        return []


@overload
def ordering() -> Callable[..., Any]:  # pragma: no cover
    ...


@overload
def ordering(
    func_or_ordering_class: Any = NOT_SET, **paginator_params: Any
) -> Callable[..., Any]:  # pragma: no cover
    ...


def ordering(
    func_or_ordering_class: Any = NOT_SET, **ordering_params: Any
) -> Callable[..., Any]:
    isfunction = inspect.isfunction(func_or_ordering_class)
    isnotset = func_or_ordering_class == NOT_SET

    ordering_class: Type[OrderingBase] = settings.ORDERING_CLASS

    if isfunction:
        return _inject_orderator(func_or_ordering_class, ordering_class)

    if not isnotset:
        ordering_class = func_or_ordering_class

    def wrapper(func: Callable[..., Any]) -> Any:
        return _inject_orderator(func, ordering_class, **ordering_params)

    return wrapper


def _inject_orderator(
    func: Callable[..., Any],
    ordering_class: Type[OrderingBase],
    **ordering_params: Any,
) -> Callable[..., Any]:
    orderator: OrderingBase = ordering_class(**ordering_params)
    orderator_kwargs_name = "ordering"
    orderator_operation_class = OrderatorOperation
    if is_async(func):
        orderator_operation_class = AsyncOrderatorOperation
    orderator_operation = orderator_operation_class(
        orderator=orderator, view_func=func, orderator_kwargs_name=orderator_kwargs_name
    )

    return orderator_operation.as_view


class OrderatorOperation:
    def __init__(
        self,
        *,
        orderator: OrderingBase,
        view_func: Callable,
        orderator_kwargs_name: str = "ordering",
    ) -> None:
        self.orderator = orderator
        self.orderator_kwargs_name = orderator_kwargs_name
        self.view_func = view_func

        orderator_view = self.get_view_function()
        self.as_view = wraps(view_func)(orderator_view)
        add_ninja_contribute_args(
            self.as_view,
            (
                self.orderator_kwargs_name,
                self.orderator.Input,
                self.orderator.InputSource,
            ),
        )
        orderator_view.orderator_operation = self  # type:ignore[attr-defined]

    @property
    def view_func_has_kwargs(self) -> bool:  # pragma: no cover
        return self.orderator.pass_parameter is not None

    def get_view_function(self) -> Callable:
        def as_view(
            request_or_controller: Union["ControllerBase", HttpRequest],
            *args: Any,
            **kw: Any,
        ) -> Any:
            func_kwargs = dict(**kw)
            ordering_params = func_kwargs.pop(self.orderator_kwargs_name)
            if self.orderator.pass_parameter:
                func_kwargs[self.orderator.pass_parameter] = ordering_params

            items = self.view_func(request_or_controller, *args, **func_kwargs)
            return self.orderator.ordering_queryset(items, ordering_params)

        return as_view


class AsyncOrderatorOperation(OrderatorOperation):
    def get_view_function(self) -> Callable:
        async def as_view(
            request_or_controller: Union["ControllerBase", HttpRequest],
            *args: Any,
            **kw: Any,
        ) -> Any:
            func_kwargs = dict(**kw)
            ordering_params = func_kwargs.pop(self.orderator_kwargs_name)
            if self.orderator.pass_parameter:
                func_kwargs[self.orderator.pass_parameter] = ordering_params

            items = await self.view_func(request_or_controller, *args, **func_kwargs)
            ordering_queryset = cast(
                Callable, sync_to_async(self.orderator.ordering_queryset)
            )
            return await ordering_queryset(items, ordering_params)

        return as_view



from ninja.pagination import LimitOffsetPagination, PageNumberPagination, PaginationBase

from ninja_extra.schemas import PaginatedResponseSchema

from .decorator import paginate
from .models import PageNumberPaginationExtra
from .operations import AsyncPaginatorOperation, PaginatorOperation

__all__ = [
    "PageNumberPagination",
    "PageNumberPaginationExtra",
    "PaginationBase",
    "LimitOffsetPagination",
    "paginate",
    "PaginatedResponseSchema",
    "PaginatorOperation",
    "AsyncPaginatorOperation",
]



from typing import Any, List, Optional, Union

from django.http import HttpResponse
from django.http.request import HttpRequest
from ninja.types import DictStrAny
from pydantic import BaseModel as PydanticModel
from pydantic import Field

from ninja_extra.types import PermissionType


class RouteContext(PydanticModel):
    """
    APIController Context which will be available to the class instance when handling request
    """

    class Config:
        arbitrary_types_allowed = True

    permission_classes: PermissionType = Field([])
    request: Union[Any, HttpRequest, None] = None
    response: Union[Any, HttpResponse, None] = None
    args: List[Any] = Field([])
    kwargs: DictStrAny = Field({})


def get_route_execution_context(
    request: HttpRequest,
    temporal_response: Optional[HttpResponse] = None,
    permission_classes: Optional[PermissionType] = None,
    *args: Any,
    **kwargs: Any,
) -> RouteContext:
    init_kwargs = {
        "permission_classes": permission_classes
        if permission_classes is not None
        else [],
        "request": request,
        "kwargs": kwargs,
        "response": temporal_response,
        "args": args,
    }
    context = RouteContext(**init_kwargs)  # type:ignore[arg-type]
    return context
